import chardet
import json
import os
import tempfile

from pygeonlp.api.metadata import Metadata


class DictionaryError(RuntimeError):
    """
    辞書データの操作の際に例外が起こると、このクラスが発生します。
    """
    pass


class Dictionary(object):
    """
    GeoNLP 辞書の管理クラス
    辞書メタデータとCSVデータに対する操作を行ないます。

    通常、開発者がこのクラスを直接操作する必要はありません。

    Attributes
    ----------
    metadata : Metadata
        メタデータのインスタンス
    csvtext : str
        CSV テキスト
    """

    def __init__(self, metadata, csvtext):
        """
        与えられたパラメータでクラス変数を初期化します。

        Parameters
        ----------
        metadata : Metadata
            メタデータのインスタンス
        csvtext : str
            CSV テキスト
        """
        self.metadata = metadata
        self.csvtext = csvtext

    @classmethod
    def download(cls, url, params=None, **kwargs):
        """
        指定された URL からウェブページをダウンロードし、
        ヘッダに記載されている json-ld を辞書メタデータとして抽出します。
        また、メタデータに書かれているデータダウンロード URL にアクセスして CSV データを取得します。
        取得したメタデータと CSV データを保持する Dictionary インスタンスを作成します。

        Parameters
        ----------
        url : str
            json-ld を含むウェブページの URL
        params : dict, optional
            requests.get に渡す params パラメータ
        **kwargs
            requests.get に渡す kwargs パラメータ

        Returns
        -------
        Dictionary
            作成した Dictionary インスタンス。
        """
        metadata = Metadata.download(url, params, **kwargs)
        csvtext = metadata.download_csv(params=params, **kwargs)

        return cls(metadata, csvtext)

    @classmethod
    def load(cls, jsonfile, csvfile):
        """
        指定したパスにある辞書メタデータ（JSONファイル）と
        地名解析辞書（CSVファイル）を読み込み Dictionary インスタンスを作成します。

        Parameters
        ----------
        jsonfile : str
            辞書メタデータファイルのパス
        csvfile : str
            地名解析辞書ファイルのパス

        Returns
        -------
        Dictionary
            作成した Dictionary インスタンス。
        """
        if not isinstance(jsonfile, str) or not isinstance(csvfile, str):
            raise TypeError("jsonfile と csvfile は str で指定してください。")

        if not os.path.exists(jsonfile):
            raise FileNotFoundError(f"jsonfile({jsonfile}) が見つかりません。")

        if not os.path.exists(csvfile):
            raise FileNotFoundError(f"csvfile({csvfile}) が見つかりません。")

        with open(jsonfile, 'r', encoding='utf-8') as f:
            jsonld = f.read()

        with open(csvfile, 'rb') as f:
            csvdata = f.read()
            encoding = chardet.detect(csvdata)
            csvtext = csvdata.decode(encoding['encoding'])

        metadata = Metadata.loads(jsonld)

        return cls(metadata, csvtext)

    @classmethod
    def create(cls, csvfile, name=None, identifier=None):
        """
        指定したパスにある地名解析辞書（CSVファイル）を読み込み
        Dictionary インスタンスを作成します。

        Parameters
        ----------
        csvfile : str
            地名解析辞書ファイルのパス
        name : str, optional
            辞書名。省略した場合、 CSV ファイルの basename を利用します。
        identifier : str, optional
            辞書 identifier。省略した場合、 CSV ファイルの basename を取り、
            ``geonlp:<basename>`` を利用します。

        Returns
        -------
        Dictionary
            作成した Dictionary インスタンス。
        """
        if not isinstance(csvfile, str):
            raise TypeError("csvfile は str で指定してください。")

        if not os.path.exists(csvfile):
            raise FileNotFoundError(f"csvfile({csvfile}) が見つかりません。")

        abspath = os.path.abspath(csvfile)
        basename = os.path.basename(csvfile)
        if basename.lower()[-4:] == '.csv':
            basename = basename[0:-4]

        if name is None:
            name = basename

        if identifier is None:
            identifier = 'geonlp:' + basename

        ld = {
            'name': name,
            'url': abspath,
            'distribution': [{
                '@type': "CopiedData",
                'contentUrl': abspath,
                'encodingFormat': 'text/csv',
            }],
            'identifier': [identifier],
        }

        with open(csvfile, 'rb') as f:
            csvdata = f.read()
            encoding = chardet.detect(csvdata)
            csvtext = csvdata.decode(encoding['encoding'])

        metadata = Metadata.loads(json.dumps(ld))

        return cls(metadata, csvtext)

    def get_identifier(self):
        """
        この辞書の identifier を返します。

        Returns
        -------
        str
            辞書 identifier 文字列。
        """
        if self.metadata is None:
            raise DictionaryError("metadata がセットされていません。")

        return self.metadata.get_identifier()

    def get_name(self):
        """
        この辞書の名前を返します。

        Returns
        -------
        str
            辞書の名前。
        """
        if self.metadata is None:
            raise DictionaryError("metadata がセットされていません。")

        return self.metadata.get_name()

    def save(self, jsonfile, csvfile):
        """
        この辞書が保持している json-ld と CSV データをファイルに保存します。

        Parameters
        ----------
        jsonfile : str
            json-ld を保存するファイル名
        csvfile : str
            CSV データを保存するファイル名

        Returns
        -------
        bool
            常に True が返ります。
            失敗した場合は例外が発生します。
        """
        if self.metadata is None:
            raise DictionaryError("metadata がセットされていません。")

        if self.csvtext is None:
            raise DictionaryError("csvtext がセットされていません。")

        with open(jsonfile, 'w', encoding='utf-8') as fp:
            fp.write(self.metadata.jsonld)

        with open(csvfile, 'w', encoding='utf-8', newline="\r\n") as fp:
            fp.write(self.csvtext)

        return True

    def add(self, capi_ma):
        """
        システムに辞書を登録します。

        Parameters
        ----------
        capi_ma : ``capi.MA``
            辞書を管理する MA オブジェクト。
            init() で初期化済みである必要があります。

        Returns
        -------
        bool
            常に True が返ります。
            失敗した場合は例外が発生します。
        """
        if self.metadata is None:
            raise DictionaryError("metadata がセットされていません。")

        if self.csvtext is None:
            raise DictionaryError("csvtext がセットされていません。")

        fp_json, fn_json = tempfile.mkstemp()
        with open(fn_json, 'w', encoding='utf-8') as fp:
            fp.write(self.metadata.jsonld)

        fp_csv, fn_csv = tempfile.mkstemp()
        with open(fn_csv, 'w', encoding='utf-8', newline="\r\n") as fp:
            fp.write(self.csvtext)

        ret = capi_ma.addDictionary(fn_json, fn_csv)
        os.remove(fn_json)
        os.remove(fn_csv)

        return ret



from pygeonlp.webapi.tests.helpers import validate_jsonrpc, write_resreq

"""
GDAL がインストールされているかどうかをチェックします
"""
"""
gdal_not_installed = False
try:
    import osgeo
except ModuleNotFoundErrlr:
    gdal_not_installed = True
"""


class TestBasicApi:
    """
    基本的な API のテスト
    """

    def test_version(self, client):
        """
        Test ``geonlp.version`` API.
        """
        import pygeonlp.api
        query = {
            'method': 'geonlp.version',
            'params': {},
            'id': 'test_version',
        }
        expected = pygeonlp.api.__version__
        result = validate_jsonrpc(client, query, expected)
        write_resreq(query, result)

    def test_parse(self, client):
        """
        Test ``geonlp.parse`` API.
        """
        query = {
            'method': 'geonlp.parse',
            'params': {'sentence': 'NIIは神保町駅から徒歩7分です。'},
            'id': 'test_parse',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 形態素解析のチェック
        words = 'NII/は/神保町駅/から/徒歩/7/分/です/。'.split('/')
        for pos, feature in enumerate(features):
            prop = feature['properties']
            assert prop['surface'] == words[pos]  # 表記が一致

            if prop['surface'] == '神保町駅':
                assert prop['node_type'] == 'GEOWORD'
            else:
                assert prop['node_type'] == 'NORMAL'

        write_resreq(query, result)

    def test_parse_geocoding(self, client):
        """
        Test ``geonlp.parse`` API using geocoding.
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': 'NIIは千代田区一ツ橋2-1-2にあります。',
                'options': {'geocoding': True}
            },
            'id': 'test_parse_geocoding',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 形態素解析のチェック
        words = 'NII/は/千代田区一ツ橋2-1-/2/に/あり/ます/。'.split('/')
        for pos, feature in enumerate(features):
            prop = feature['properties']
            assert prop['surface'] == words[pos]  # 表記が一致

            if prop['surface'] == '千代田区一ツ橋2-1-':
                assert prop['node_type'] == 'ADDRESS'
            else:
                assert prop['node_type'] == 'NORMAL'

        write_resreq(query, result)

    def test_parse_without_geocoding(self, client):
        """
        Test ``geonlp.parse`` API using geocoding.
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': 'NIIは千代田区一ツ橋2-1-2にあります。',
                'options': {'geocoding': False}
            },
            'id': 'test_parse_without_geocoding',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 形態素解析のチェック
        words = 'NII/は/千代田区/一ツ橋/2/-/1/-/2/に/あり/ます/。'.split('/')
        for pos, feature in enumerate(features):
            prop = feature['properties']
            assert prop['surface'] == words[pos]  # 表記が一致

            if prop['surface'] == '千代田区':
                assert prop['node_type'] == 'GEOWORD'
            elif prop['surface'] == '一ツ橋':
                assert prop['node_type'] == 'NORMAL'
            else:
                assert prop['node_type'] == 'NORMAL'

    def test_parseStructured(self, client):
        """
        Test ``geonlp.parseStructured`` API.
        """
        query = {
            'method': 'geonlp.parseStructured',
            'params': {
                'sentence_list': [
                    'NIIは神保町駅から徒歩7分です。',
                    '千代田区一ツ橋2-1-2にあります。',
                    '竹橋駅も近いです。', ]
            },
            'id': 'test_parseStructured',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 形態素解析のチェック
        words = 'NII/は/神保町駅/から/徒歩/7/分/です/。'.split('/') +\
            '千代田区/一ツ橋/2/-/1/-/2/に/あり/ます/。'.split('/') +\
            '竹橋駅/も/近い/です/。'.split('/')
        for pos, feature in enumerate(features):
            prop = feature['properties']
            assert prop['surface'] == words[pos]  # 表記が一致

            if prop['surface'] in ('神保町駅', '千代田区', '竹橋駅'):
                assert prop['node_type'] == 'GEOWORD'
            else:
                assert prop['node_type'] == 'NORMAL'

        write_resreq(query, result)

    def test_search(self, client):
        """
        Test ``geonlp.search`` API.
        """
        query = {
            'method': 'geonlp.search',
            'params': {'key': '国会議事堂前'},
            'id': 'test_search',
        }
        expected = {
            'Bn4q6d': {
                'body': '国会議事堂前', 'dictionary_id': 3,
                'dictionary_identifier': 'geonlp:ksj-station-N02',
                'entry_id': 'LrGGxY', 'geolod_id': 'Bn4q6d',
                'hypernym': ['東京地下鉄', '4号線丸ノ内線'],
                'institution_type': '民営鉄道',
                'latitude': '35.674845',
                'longitude': '139.74534166666666',
                'ne_class': '鉄道施設/鉄道駅',
                'railway_class': '普通鉄道',
                'suffix': ['駅', '']
            },
            'cE8W4w': {
                'body': '国会議事堂前', 'dictionary_id': 3,
                'dictionary_identifier': 'geonlp:ksj-station-N02',
                'entry_id': '4NFELa',
                'geolod_id': 'cE8W4w',
                'hypernym': ['東京地下鉄', '9号線千代田線'],
                'institution_type': '民営鉄道',
                'latitude': '35.673543333333335',
                'longitude': '139.74305333333334',
                'ne_class': '鉄道施設/鉄道駅',
                'railway_class': '普通鉄道',
                'suffix': ['駅', '']
            }
        }
        result = validate_jsonrpc(client, query, expected)
        write_resreq(query, result)

    def test_getGeoInfo_idlist(self, client):
        """
        Test ``geonlp.getGeoInfo`` API with multiple geolod_id.
        """
        query = {
            'method': 'geonlp.getGeoInfo',
            'params': {'idlist': ['Bn4q6d', 'cE8W4w']},
            'id': 'test_getGeoInfo_idlist',
        }
        expected = {
            'Bn4q6d': {
                'body': '国会議事堂前', 'dictionary_id': 3,
                'dictionary_identifier': 'geonlp:ksj-station-N02',
                'entry_id': 'LrGGxY', 'geolod_id': 'Bn4q6d',
                'hypernym': ['東京地下鉄', '4号線丸ノ内線'],
                'institution_type': '民営鉄道',
                'latitude': '35.674845',
                'longitude': '139.74534166666666',
                'ne_class': '鉄道施設/鉄道駅',
                'railway_class': '普通鉄道',
                'suffix': ['駅', '']
            },
            'cE8W4w': {
                'body': '国会議事堂前', 'dictionary_id': 3,
                'dictionary_identifier': 'geonlp:ksj-station-N02',
                'entry_id': '4NFELa',
                'geolod_id': 'cE8W4w',
                'hypernym': ['東京地下鉄', '9号線千代田線'],
                'institution_type': '民営鉄道',
                'latitude': '35.673543333333335',
                'longitude': '139.74305333333334',
                'ne_class': '鉄道施設/鉄道駅',
                'railway_class': '普通鉄道',
                'suffix': ['駅', '']
            }
        }
        result = validate_jsonrpc(client, query, expected)
        write_resreq(query, result)

    def test_getGeoInfo_nocandidate(self, client):
        """
        Test ``geonlp.getGeoInfo`` API with invalid geolod_id.
        """
        query = {
            'method': 'geonlp.getGeoInfo',
            'params': {'idlist': ['aaaaaa']},
            'id': 'test_getGeoInfo_nocandidate',
        }
        expected = {'aaaaaa': None}
        result = validate_jsonrpc(client, query, expected)
        write_resreq(query, result)

    def test_getDictionaries(self, client):
        """
        Test ``geonlp.getDictionaries`` API.
        """
        query = {
            'method': 'geonlp.getDictionaries',
            'params': {},
            'id': 'test_getDictionaries',
        }
        expected = ['geonlp:geoshape-city',
                    'geonlp:geoshape-pref',
                    'geonlp:ksj-station-N02']
        result = validate_jsonrpc(client, query, expected)
        write_resreq(query, result)

    def test_getDictionaryInfo(self, client):
        """
        Test ``geonlp.getDictionaryInfo`` API.
        """
        query = {
            'method': 'geonlp.getDictionaryInfo',
            'params': {'identifier': 'geonlp:ksj-station-N02'},
            'id': 'test_getDictionaryInfo',
        }
        expected = '{"@context": "https://schema.org/", "@type": "Dataset", "alternateName": "", "creator": [{"@type": "Organization", "name": "株式会社情報試作室", "sameAs": "https://www.info-proto.com/"}], "dateModified": "2021-08-27T17:18:18+09:00", "description": "国土数値情報「鉄道データ（N02）」から作成した、日本の鉄道駅（地下鉄を含む）の辞書です。hypernym には運営者名と路線名を記載しています。「都営」ではなく「東京都」のようになっていますので注意してください。自由フィールドとして、railway_classに「鉄道区分」、institution_typeに「事業者種別」を含みます。", "distribution": [{"@type": "DataDownload", "contentUrl": "https://www.info-proto.com/static/ksj-station-N02.csv", "encodingFormat": "text/csv"}], "identifier": ["geonlp:ksj-station-N02"], "isBasedOn": {"@type": "CreativeWork", "name": "国土数値情報 鉄道データ", "url": "https://nlftp.mlit.go.jp/ksj/gml/datalist/KsjTmplt-N02-v2_2.html"}, "keywords": ["GeoNLP", "地名辞書"], "license": "https://creativecommons.org/licenses/by/4.0/", "name": "国土数値情報 鉄道データ（駅）", "size": "10191", "spatialCoverage": {"@type": "Place", "geo": {"@type": "GeoShape", "box": "26.193265 127.652285 45.41616333333333 145.59723"}}, "temporalCoverage": "../..", "url": "https://www.info-proto.com/static/ksj-station-N02.html"}'
        result = validate_jsonrpc(client, query, expected)
        write_resreq(query, result)

    def test_getDictionaryInfo_nocandidate(self, client):
        """
        Test ``geonlp.getDictionaryInfo`` API with invalid identifier.
        """
        query = {
            'method': 'geonlp.getDictionaryInfo',
            'params': {'identifier': 'geonlp:abcdef'},
            'id': 'test_getDictionaryInfo_nocandidate',
        }
        expected = None
        result = validate_jsonrpc(client, query, expected)
        write_resreq(query, result)

    def test_addressGeocoding(self, client):
        """
        Test ``geonlp.addressGeocoding`` API.
        """
        query = {
            'method': 'geonlp.addressGeocoding',
            'params': {'address': '千代田区一ツ橋2-1-2'},
            'id': 'test_addressGeocoding',
        }
        expected = "*"
        result = validate_jsonrpc(client, query, expected)
        write_resreq(query, result)


class TestParseOptions:
    """
    Parse options のテスト
    """

    def test_parse_option_set_dic(self, client):
        """
        Test 'set-dic' option.

        Notes
        -----
        'set-dic' に r'geoshape' を指定すると、
        identifier に 'geoshape' を含む辞書だけが利用されます。

        'geoshape-city' が利用できるため、
        「和歌山市（市）」に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '和歌山市は晴れ。',
                'options': {'set-dic': r'geoshape'}
            },
            'id': 'test_parse_set_dic',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '和歌山市'
        assert prop['geoword_properties']['geolod_id'] == 'lQccqK'

        write_resreq(query, result)

    def test_parse_option_set_dic_list(self, client):
        """
        Test 'set-dic' option with list of identifiers.

        Notes
        -----
        'set-dic' に ['geoshape'] を指定すると、
        identifier が 'geoshape' と文字列として等しい辞書だけが
        利用されます。

        一致する辞書は無いため、「和歌山」に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '和歌山市は晴れ。',
                'options': {'set-dic': ['geoshape']}
            },
            'id': 'test_parse_set_dic_list',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '和歌山'
        assert 'geoword_properties' not in prop

    def test_parse_option_remove_dic(self, client):
        """
        Test 'remove-dic' option.

        Notes
        -----
        'remove-dic' に r'station' を指定すると、
        identifier に 'station' を含む辞書は利用されません。

        'geoshape-city' が利用できるため、
        「和歌山市（市）」に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '和歌山市は晴れ。',
                'options': {'remove-dic': r'station'}
            },
            'id': 'test_parse_remove_dic',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '和歌山市'
        assert prop['geoword_properties']['dictionary_identifier'] == \
            'geonlp:geoshape-city'

        write_resreq(query, result)

    def test_parse_option_add_dic(self, client):
        """
        Test 'add-dic' option.

        Notes
        -----
        'add-dic' に r'city' を指定すると、
        'remove-dic' で除外指定されていても、
        identifier に 'city' を含む辞書は利用されます。

        'geoshape-city' は 'remove-dic' で除外されますが
        'add-dic' で追加されるため利用可能になり、
        「和歌山市（市）」に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '和歌山市は晴れ。',
                'options': {'remove-dic': r'.*',
                            'add-dic': r'city'}
            },
            'id': 'test_parse_add_dic',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '和歌山市'
        assert prop['geoword_properties']['dictionary_identifier'] == \
            'geonlp:geoshape-city'

        write_resreq(query, result)

    def test_parse_option_set_class(self, client):
        """
        Test 'set-class' option.

        Notes
        -----
        'set-class' に [r'.*', r'-鉄道施設/.*'] を指定すると、
        全ての固有名クラスから '鉄道施設/.*' を除いたものを持つ
        地名語を候補として利用します。

        「和歌山市（駅）」は鉄道施設なので除外され、
        「和歌山市（市）」に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '和歌山市は晴れ。',
                'options': {'set-class': [r'.*', r'-鉄道施設/.*']}
            },
            'id': 'test_parse_set_class',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '和歌山市'
        assert prop['geoword_properties']['dictionary_identifier'] == \
            'geonlp:geoshape-city'

        write_resreq(query, result)

    def test_parse_option_set_class_except(self, client):
        """
        Test 'set-class' option.

        Notes
        -----
        'set-class' に [r'.*', r'-鉄道施設/.*', r'鉄道施設/鉄道駅'] を
        指定すると、この順番に評価するため「鉄道駅」は利用されます。

        「和歌山市（駅）」は鉄道施設/鉄道駅なので対象になります。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '和歌山市駅は晴れ。',
                'options': {'set-class': [
                    r'.*', r'-鉄道施設/.*', r'鉄道施設/鉄道駅'
                ]}
            },
            'id': 'test_parse_set_class_except',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '和歌山市駅'
        assert prop['geoword_properties']['dictionary_identifier'] == \
            'geonlp:ksj-station-N02'

    def test_parse_option_add_class(self, client):
        """
        Test 'add-class' option.

        Notes
        -----
        'add-class' で利用するクラスのリストを指定できます。

        'add-class' に [r'-.*', '市区町村'] を指定すると、
        全ての固有名クラスを対象クラスから除外し、
        次に「市区町村」クラスの地名語を対象に加えます。

        「和歌山市（駅）」は鉄道施設なので除外され、
        「和歌山市（市）」に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '和歌山市は晴れ。',
                'options': {'add-class': [
                    r'-.*', '市区町村'
                ]}
            },
            'id': 'test_parse_add_class',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '和歌山市'
        assert prop['geoword_properties']['dictionary_identifier'] == \
            'geonlp:geoshape-city'

        write_resreq(query, result)

    def test_parse_option_remove_class(self, client):
        """
        Test 'remove-class' option.

        Notes
        -----
        'set-class' で一度に指定する代わりに、
        'remove-class' で除外するクラスのリストを指定できます。

        'remove-class' に [r'.*', '-市区町村'] を指定すると、
        全ての固有名クラスを対象クラスから除外し、
        次に「市区町村」クラスの地名語を対象に加えます。

        'remove-class' のリストではクラス名の先頭に '-' を付けると
        追加の意味になります。

        「和歌山市（駅）」は鉄道施設なので除外され、
        「和歌山市（市）」に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '和歌山市は晴れ。',
                'options': {'remove-class': [
                    r'.*', '-市区町村'
                ]}
            },
            'id': 'test_parse_remove_class',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '和歌山市'
        assert prop['geoword_properties']['dictionary_identifier'] == \
            'geonlp:geoshape-city'

        write_resreq(query, result)

    def test_parse_option_add_remove_class(self, client):
        """
        Test 'add-class' + 'remove-class' options.

        Notes
        -----
        'set-class' で一度に指定する代わりに、
        'add-class' で利用するクラスのリストを、
        'remove-class' で除外するクラスのリストを指定できます。

        ただし必ず 'remove-class' が先に評価されます。

        'add-class' で ['市区町村'] を、
        'remove-class' で [r'.*'] を指定すると、
        全ての固有名クラスを対象クラスから除外し、
        次に「市区町村」クラスの地名語を対象に加えます。

        「和歌山市（駅）」は鉄道施設なので除外され、
        「和歌山市（市）」に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '和歌山市は晴れ。',
                'options': {'add-class': ['市区町村'],
                            'remove-class': [r'.*']}
            },
            'id': 'test_parse_add_remove_class',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '和歌山市'
        assert prop['geoword_properties']['dictionary_identifier'] == \
            'geonlp:geoshape-city'


class TestSpatialFilters:
    """
    SpatialFilter のテスト
    """

    def test_parse_filter_geo_contains(self, client, require_gdal):
        """
        Test 'geo-contains' filter.

        Notes
        -----
        'geo-contains' に GeoJSON または GeoJSON を返す URL を
        指定すると、その範囲に含まれる地名語だけが候補になります。

        東京都付近の四角形を空間範囲として指定することで
        東京都の府中駅に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '府中に行きます。',
                'options': {
                    'geo-contains': {
                        "type": "Polygon",
                        "coordinates": [
                            [[139.43, 35.54], [139.91, 35.54],
                             [139.91, 35.83], [139.43, 35.83],
                                [139.43, 35.54]]]
                    }}
            },
            'id': 'test_parse_geo_contains',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '府中'
        assert '京王線' in prop['geoword_properties']['hypernym']

        write_resreq(query, result)

    def test_parse_filter_geo_disjoint(self, client, require_gdal):
        """
        Test 'geo-disjoint' filter.

        Notes
        -----
        'geo-disjoint' に GeoJSON または GeoJSON を返す URL を
        指定すると、その範囲に含まれない地名語だけが候補になります。

        東京都付近の四角形を空間範囲として指定することで
        京王線の府中駅以外に解決されます。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': '府中に行きます。',
                'options': {
                    'geo-disjoint': 'https://geoshape.ex.nii.ac.jp/city/geojson/20230101/13/13206A1968.geojson'
                }
            },
            'id': 'test_parse_geo_disjoint',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        prop = features[0]['properties']
        assert prop['surface'] == '府中'
        assert '京王線' not in prop['geoword_properties']['hypernym']

        write_resreq(query, result)


class TestTemporalFilters:
    """
    TemporalFilter のテスト
    """

    sentence = ('田無市と保谷市は2001年1月21日に合併して'
                '西東京市になりました。')

    def test_time_exists(self, client):
        """
        Test 'time-exists' filter.

        Notes
        -----
        'time-exists' に日時または日時のペアを指定すると、
        その時点・期間に存在した地名語だけが候補になります。

        西東京市は指定した期間内に設置されたので
        田無市と保谷市、西東京市は全て地名語になります。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': self.sentence,
                'options': {
                    'time-exists': ['2000-01-01', '2001-02-01']
                }
            },
            'id': 'test_time_exists',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        for feature in features:
            prop = feature['properties']
            if prop['surface'] in ('田無市', '保谷市', '西東京市',):
                assert prop['node_type'] == 'GEOWORD'
            else:
                assert prop['node_type'] == 'NORMAL'

        write_resreq(query, result)

    def test_time_exists_single_value(self, client):
        """
        Test 'time-exists' filter with single value.

        Notes
        -----
        'time-exists' に日時または日時を1つ指定すると、
        その時点に存在した地名語だけが候補になります。

        西東京市は指定した時点にはまだ設置されていないので
        田無市と保谷市が地名語、西東京市は固有名詞になります。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': self.sentence,
                'options': {
                    'time-exists': '2000-01-01'
                }
            },
            'id': 'test_time_exists',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        for feature in features:
            prop = feature['properties']
            if prop['surface'] in ('田無市', '保谷市',):
                assert prop['node_type'] == 'GEOWORD'
            elif prop['surface'] in ('西東京市',):
                assert prop['node_type'] == 'NORMAL'
            else:
                assert prop['node_type'] == 'NORMAL'

    def test_time_exists_dict(self, client):
        """
        Test 'time-exists' filter with dict values.

        Notes
        -----
        'time-exists' に 'date_from' と 'date_to' を持つ
        dict を指定すると、
        その期間に存在した地名語だけが候補になります。

        西東京市は指定した期間内に設置されたので
        田無市と保谷市、西東京市は全て地名語になります。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': self.sentence,
                'options': {
                    'time-exists': {
                        'date_from': '2000-01-01',
                        'date_to': '2001-02-01'
                    }
                }
            },
            'id': 'test_time_exists_dict',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        for feature in features:
            prop = feature['properties']
            if prop['surface'] in ('田無市', '保谷市', '西東京市',):
                assert prop['node_type'] == 'GEOWORD'
            else:
                assert prop['node_type'] == 'NORMAL'

    def test_time_before(self, client):
        """
        Test 'time-before' filter.

        Notes
        -----
        'time-before' に日時または日時のペアを指定すると、
        その時点・期間の開始時より前に存在した地名語だけが
        候補になります。

        西東京市は指定した期間の開始時より後に設置されたので
        田無市と保谷市が地名語、西東京市は固有名詞になります。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': self.sentence,
                'options': {
                    'time-before': ['2000-01-01', '2001-02-01']
                }
            },
            'id': 'test_time_before',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        for feature in features:
            prop = feature['properties']
            if prop['surface'] in ('田無市', '保谷市',):
                assert prop['node_type'] == 'GEOWORD'
            elif prop['surface'] in ('西東京市',):
                assert prop['node_type'] == 'NORMAL'
            else:
                assert prop['node_type'] == 'NORMAL'

        write_resreq(query, result)

    def test_time_after(self, client):
        """
        Test 'time-after' filter.

        Notes
        -----
        'time-after' に日時または日時のペアを指定すると、
        その時点・期間の終了時より後に存在した地名語だけが
        候補になります。

        田無市と保谷市は指定した期間の終了時より前に廃止されたので
        田無市と保谷市は固有名詞、西東京市は地名語になります。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': self.sentence,
                'options': {
                    'time-after': ['2000-01-01', '2001-02-01']
                }
            },
            'id': 'test_time_after',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        for feature in features:
            prop = feature['properties']
            if prop['surface'] in ('西東京市',):
                assert prop['node_type'] == 'GEOWORD'
            elif prop['surface'] in ('田無市', '保谷市',):
                assert prop['node_type'] == 'NORMAL'
            else:
                assert prop['node_type'] == 'NORMAL'

        write_resreq(query, result)

    def test_time_overlaps(self, client):
        """
        Test 'time-overlaps' filter.

        Notes
        -----
        'time-overlaps' に日時または日時のペアを指定すると、
        その時点・期間の終了時より後に存在した地名語だけが
        候補になります。

        西東京市は指定した期間内に設置されたので
        田無市と保谷市、西東京市は全て地名語になります。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': self.sentence,
                'options': {
                    'time-overlaps': ['2000-01-01', '2001-02-01']
                }
            },
            'id': 'test_time_overlaps',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        for feature in features:
            prop = feature['properties']
            if prop['surface'] in ('田無市', '保谷市', '西東京市',):
                assert prop['node_type'] == 'GEOWORD'
            else:
                assert prop['node_type'] == 'NORMAL'

        write_resreq(query, result)

    def test_time_covers(self, client):
        """
        Test 'time-covers' filter.

        Notes
        -----
        'time-covers' に日時または日時のペアを指定すると、
        その期間の開始時より終了時まで存在し続けた地名語だけが
        候補になります。

        西東京市は指定した期間内に設置され、
        田無市と保谷市は期間内に廃止されたので、
        田無市、保谷市、西東京市は全て固有名詞になります。
        """
        query = {
            'method': 'geonlp.parse',
            'params': {
                'sentence': self.sentence,
                'options': {
                    'time-covers': ['2000-01-01', '2001-02-01']
                }
            },
            'id': 'test_time_covers',
        }
        expected = '*'
        result = validate_jsonrpc(client, query, expected)

        # GeoJSON Feature Collection のチェック
        assert result['type'] == 'FeatureCollection'
        assert 'features' in result
        features = result['features']

        # 地名語のチェック
        for feature in features:
            prop = feature['properties']
            assert prop['node_type'] == 'NORMAL'

        write_resreq(query, result)



import logging
import os
import unittest

import jageocoder

import pygeonlp.api as api
from pygeonlp.api.node import Node
from pygeonlp.api.workflow import Workflow


logger = logging.getLogger(__name__)

"""
jageocoder を利用するテスト。

このモジュールのテストを行なうには、次のコマンドを実行してください。

python -m unittest -v pygeonlp.tests.test_api_geocoder
"""

# os.environ["JAGEOCODER_SERVER_URL"] = "https://jageocoder.info-proto.com/jsonrpc"
# os.environ["JAGEOCODER_DB2_DIR"] = "/home/user/jageocoder/db2"


class TestModuleMethods(unittest.TestCase):

    workflow = None

    @classmethod
    def setUpClass(cls):
        # Create test database
        testdir = os.path.abspath(os.path.join(os.getcwd(), 'apitest'))
        os.environ['GEONLP_DB_DIR'] = testdir
        os.makedirs(testdir, 0o755, exist_ok=True)
        dict_manager = api.dict_manager.DictManager(db_dir=testdir)
        dict_manager.setupBasicDatabase()

        # Initialize jageocoder
        try:
            jageocoder.init()
            cls.workflow = Workflow(
                db_dir=testdir,
                jageocoder=jageocoder.get_module_tree(),
            )
            cls.parser = cls.workflow.parser
        except jageocoder.exceptions.JageocoderError:
            cls.workflow = None

    def setUp(self):
        if self.workflow is None:
            self.skipTest('住所辞書がインストールされていません。')

    def test_geoparse_with_address(self):
        """
        住所が正しく解析できることを確認する。
        """
        lattice = self.parser.analyze_sentence('千代田区一ツ橋2-1-2')
        lattice_address = self.parser.add_address_candidates(lattice)
        self.assertIsInstance(lattice_address, list)
        self.assertEqual(len(lattice_address), 2)
        self.assertEqual(len(lattice_address[0]), 1)
        self.assertIsInstance(lattice_address[0][0], Node)
        self.assertEqual(lattice_address[0][0].surface, '千代田区一ツ橋2-1-')

        # 内部の形態素を確認
        answer = ['千代田区', '一ツ橋', '2', '-', '1', '-']
        inner_morphemes = lattice_address[0][0].morphemes
        self.assertEqual(len(inner_morphemes), len(answer))
        for i, node in enumerate(inner_morphemes):
            self.assertIsInstance(node, Node)
            self.assertEqual(node.surface, answer[i])

    def test_geoparse_with_address_no_cutoff(self):
        """
        住所の解析の際に後方の語を途中で切断しないことを確認する。
        """
        lattice = self.parser.analyze_sentence('喜多方市三島町')
        lattice_address = self.parser.add_address_candidates(lattice)
        self.assertIsInstance(lattice_address, list)
        self.assertEqual(len(lattice_address), 2)
        self.assertEqual(lattice_address[0][0].surface, '喜多方市')
        self.assertEqual(lattice_address[1][0].surface, '三島町')

    def test_geoparse_location_disambiguation_hypernym(self):
        """
        住所内の地名語に複数の候補があるとき、
        上位の地名から正しい候補を選択できることを確認する。
        """
        lattice = self.parser.analyze_sentence('横浜市緑区寺山町118番地')
        lattice_address = self.parser.add_address_candidates(lattice)
        self.assertIsInstance(lattice_address, list)
        self.assertEqual(len(lattice_address), 1)
        self.assertEqual(len(lattice_address[0]), 1)
        self.assertIsInstance(lattice_address[0][0], Node)
        self.assertEqual(lattice_address[0][0].surface, '横浜市緑区寺山町118番地')

        # 内部の形態素を確認
        # 「寺山町」が MeCab の結果通り、住所要素の区切りと一致しない
        answer = ['横浜市', '緑区', '寺山', '町', '118', '番地']
        inner_morphemes = lattice_address[0][0].morphemes
        self.assertEqual(len(inner_morphemes), len(answer))
        for i, node in enumerate(inner_morphemes):
            self.assertIsInstance(node, Node)
            self.assertEqual(node.surface, answer[i])

        # '緑区' が '横浜市緑区' に解決されていることを確認
        midoriku = inner_morphemes[1]
        self.assertIn('横浜市', midoriku.prop['hypernym'])

    def test_geoparse_location_disambiguation_hyponym(self):
        """
        住所内の地名語に複数の候補があるとき、
        下位の地名から正しい候補を選択できることを確認する。
        """
        lattice = self.parser.analyze_sentence('緑区おゆみ野三丁目15番地3')
        lattice_address = self.parser.add_address_candidates(lattice)
        self.assertIsInstance(lattice_address, list)
        self.assertEqual(len(lattice_address), 2)
        self.assertEqual(len(lattice_address[0]), 1)
        self.assertIsInstance(lattice_address[0][0], Node)
        self.assertEqual(lattice_address[0][0].surface, '緑区おゆみ野三丁目15番地')

        # '緑区' が '千葉市緑区' に解決されていることを確認
        inner_morphemes = lattice_address[0][0].morphemes
        midoriku = inner_morphemes[0]
        self.assertIn('千葉市', midoriku.prop['hypernym'])

    def test_geoparse_result(self):
        """
        住所を含む文の解析を geoparse() で実行した結果を確認する。
        """
        answer = [
            {"surface": "NII", "node_type": "NORMAL"},
            {"surface": "は", "node_type": "NORMAL"},
            {"surface": "千代田区一ツ橋2-1-", "node_type": "ADDRESS"},
            {"surface": "2", "node_type": "NORMAL"},
            {"surface": "に", "node_type": "NORMAL"},
            {"surface": "あり", "node_type": "NORMAL"},
            {"surface": "ます", "node_type": "NORMAL"},
            {"surface": "。", "node_type": "NORMAL"},
        ]
        result = self.workflow.geoparse('NIIは千代田区一ツ橋2-1-2にあります。')
        for i, feature in enumerate(result):
            prop = feature['properties']
            self.assertEqual(prop['surface'], answer[i]['surface'])
            self.assertEqual(prop['node_type'], answer[i]['node_type'])

    def test_geoparse_with_address(self):
        """
        住所要素の親子関係を利用して要素を選択できることを確認する。
        """
        lattice = self.parser.analyze_sentence('静岡県浜松市天竜区水窪町山住')
        lattice_address = self.parser.add_address_candidates(lattice)

        # 内部の形態素を確認
        answer = ['静岡県', '浜松市', '天竜区', '水窪町', '山住']
        inner_morphemes = lattice_address[0][0].morphemes
        self.assertEqual(len(inner_morphemes), len(answer))
        self.assertEqual(inner_morphemes[0].node_type, Node.GEOWORD)
        self.assertEqual(inner_morphemes[1].node_type, Node.GEOWORD)
        self.assertEqual(inner_morphemes[2].node_type, Node.GEOWORD)
        for i, node in enumerate(inner_morphemes):
            self.assertIsInstance(node, Node)
            self.assertEqual(node.surface, answer[i])



import json
import sys


def print_response(rv):
    print("code:{}, data:'{}'".format(rv.status_code, rv.data),
          file=sys.stderr)


def call_jsonrpc(client, query):
    """
    Send query and return the response.

    Parameters
    ----------
    client: client fixture
    query: dict
        The query string to POST to the server
        in decoded JSON format.
    Return
    ------
    any
        The JSON decoded value of the response.
    """
    if isinstance(query, dict):
        query = json.dumps(query)

    headers = {"Content-Type": "application/json"}
    rv = client.post('/api', data=query, headers=headers)
    assert rv.status_code == 200
    response_object = json.loads(rv.data)
    return response_object


def validate_jsonrpc(client, query, expected):
    """
    Send query and compare the result with expected.

    Parameters
    ----------
    client: client fixture
    query: dict
        The query string to POST to the server
        in decoded JSON format.
    expected: any
        The value expected as a result.
        The JSON decoded value stored in the 'result'
        field of the response is used.

        When '*' is passed as 'expected', any response will pass.

    Return
    ------
    any
        The JSON decoded value stored in the 'result'
        field of the response.
    """
    response_object = call_jsonrpc(client, query)
    result = response_object['result']
    if expected != '*':
        if result != expected:
            print("result: '{}'".format(result), file=sys.stderr)

        assert result == expected

    return result


def validate_jsonrpc_error(client, query, expected_code):
    """
    Send query and compare the status code with expected_code.

    Parameters
    ----------
    client: client fixture
    query: dict
        The query object to POST to the server
        in decoded JSON format.
    expected_code: int
        The expected status_code.
    """
    if isinstance(query, dict):
        query = json.dumps(query)

    headers = {"Content-Type": "application/json"}
    rv = client.post('/api', data=query, headers=headers)
    assert rv.status_code == expected_code


def write_resreq(request, response):
    """
    Write request and response JSON into files.

    Notes
    -----
    The request JSON will be written to
    'docs/source/json/{request['id']}_req.json', and
    the response to 'docs/source/json/{request['id']}_res.json'.

    Parameters
    ----------
    request: dict
        The request in decoded JSON format.
    response: dict
        The response in decoded JSON format.
    """
    basename = request['id']
    with open("docs/source/json/{}_req.json".format(basename), "w") as f:
        json.dump(request, f, indent=2, ensure_ascii=False)
    with open("docs/source/json/{}_res.json".format(basename), "w") as f:
        json.dump(response, f, indent=2, ensure_ascii=False)



"""
このモジュールは、特定の地域内の地名語だけを抽出するといった
空間フィルタを実装しています。

空間演算に OSGeo の GDAL パッケージが必要です。
GDAL Python パッケージをインストールするには、利用している OS で
インストールした libgdal のバージョンに合わせる必要があります。
詳しくは https://pypi.org/project/GDAL/#history を確認してください。

GDAL パッケージをインストールしない場合は空間フィルタは利用できません。
"""

import json
from logging import getLogger
import urllib

from pygeonlp.api.filter import Filter, FilterError

try:
    from osgeo import ogr
except ModuleNotFoundError:
    raise FilterError((
        "SpatialFilter を利用するには 'gdal' パッケージを"
        "インストールしてください。"))

logger = getLogger(__name__)


class SpatialFilter(Filter):
    """
    空間フィルタの基底クラス。

    空間フィルタは、形態素に対応する地名語ノード候補と住所ノード候補のうち、
    条件に一致するものを残し、一致しないものを削除します。
    全ての候補が条件に一致しない場合、品詞情報として固有名詞を持つ
    一般ノードを一つ作成して置き換えます。

    空間フィルタでは、地名語ノードと住所ノードの geometry が持つ GeoJSON を
    そのノードの地点とします。
    """

    def __init__(self, geojson_or_url, **kwargs):
        """
        フィルタを初期化します。

        Parameters
        ----------
        geojson_or_url : str
            空間範囲を表す GeoJSON　または GeoJSON ファイルを取得できる URL。
        """
        super().__init__()
        self.when_all_failed = 'convert_to_normal'
        self.geo = self.__class__.get_geometry(geojson_or_url)

    @classmethod
    def get_geometry_from_geojson_url(cls, url):
        """
        指定された URL から GeoJSON を含むファイルをダウンロードして、
        osgeo.ogr.Geometry オブジェクトを作成します。

        Parameters
        ----------
        url : str
            対象ファイルの URL。

        Returns
        -------
        osgeo.ogr.Geometry
            GeoJSON から作成した Geometry オブジェクト。

        Examples
        --------
        >>> from pygeonlp.api.spatial_filter import SpatialFilter
        >>> SpatialFilter.get_geometry_from_geojson_url(
        ...   'https://geoshape.ex.nii.ac.jp/city/geojson/20200101/13/13101A1968.geojson').ExportToWkt()
        'MULTIPOLYGON (((139.73150362 35.68150121,139.73119903 35.68198095,...,139.73150362 35.68150121)))'
        """
        geo = None

        try:
            req = urllib.request.Request(url)
            with urllib.request.urlopen(req) as res:
                geojson = res.read().decode('utf-8')

            geo = cls.get_geometry_from_geojson(geojson)

        except RuntimeError as e:
            raise FilterError(str(e))

        return geo

    @classmethod
    def get_geometry_from_geojson(cls, geojson):
        """
        指定された GeoJSON 文字列から osgeo.ogr.Geometry オブジェクトを作成します。

        Parameters
        ----------
        geojson : str
            GeoJSON 文字列。
            GeoJSON をデコードした dict も利用できます。

        Returns
        -------
        osgeo.ogr.Geometry
            GeoJSON から作成した Geometry オブジェクト。

        Examples
        --------
        >>> from pygeonlp.api.spatial_filter import SpatialFilter
        >>> SpatialFilter.get_geometry_from_geojson(
        ...   {'type':'Point','coordinates':[139.6917337, 35.6895014]}
        ... ).ExportToWkt()
        'POINT (139.6917337 35.6895014)'
        """
        if isinstance(geojson, str):
            geoobj = json.loads(geojson)
        else:
            geoobj = geojson

        if geoobj['type'] == 'FeatureCollection':
            geojson = json.dumps(geoobj['features'][0]['geometry'])
        elif geoobj['type'] == 'Feature':
            geojson = json.dumps(geoobj['geometry'])
        elif 'coordinates' in geoobj:  # geometry type
            geojson = json.dumps(geoobj)
        else:
            raise FilterError('geojson の種別を判定できませんでした: ' + geojson)

        geo = ogr.CreateGeometryFromJson(geojson)

        if not geo:
            raise FilterError('Cannot parse the given geojson: ' + geojson)

        return geo

    @classmethod
    def get_geometry(cls, geojson_or_url):
        """
        GeoJSON または GeoJSON ファイルを取得できる URL から、
        osgeo.ogr.Geometry オブジェクトを作成します。

        Parameters
        ----------
        geojson_or_url : str
            GeoJSON 文字列またはデコードした dict、
            または GeoJSON ファイルをダウンロードできる URL。

        Returns
        -------
        osgeo.ogr.Geometry
            GeoJSON から作成した Geometry オブジェクト。

        Examples
        --------
        >>> from pygeonlp.api.spatial_filter import SpatialFilter
        >>> SpatialFilter.get_geometry({'type':'Point','coordinates':[139.6917337, 35.6895014]}).ExportToWkt()
        'POINT (139.6917337 35.6895014)'
        >>> SpatialFilter.get_geometry('https://geoshape.ex.nii.ac.jp/city/geojson/20200101/13/13101A1968.geojson').ExportToWkt()
        'MULTIPOLYGON (((139.73150362 35.68150121,139.73119903 35.68198095,...,139.73150362 35.68150121)))'
        """
        if isinstance(geojson_or_url, str):
            try:
                geojson = json.loads(geojson_or_url)
                geo = cls.get_geometry_from_geojson(geojson)
                return geo
            except json.decoder.JSONDecodeError:
                geo = cls.get_geometry_from_geojson_url(geojson_or_url)
        else:
            geo = cls.get_geometry_from_geojson(geojson_or_url)

        return geo

    @classmethod
    def point_from_candidate(cls, candidate):
        """
        候補ノードにふくまれる経度緯度座標から ogr.Geometry オブジェクトを作成します。
        経度緯度が含まれていない場合は None を返します。

        Parameters
        ----------
        candidate : pygeonlp.api.node.Node
            候補ノード。

        Returns
        -------
        osgeo.ogr.Geometry
            Point タイプの Geometry オブジェクト。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> from pygeonlp.api.spatial_filter import SpatialFilter
        >>> api.init()
        >>> SpatialFilter.point_from_candidate(api.analyze('国会議事堂前')[0][0]).ExportToWkt()
        'POINT (139.745341666667 35.674845)'
        """

        if candidate.geometry is None:
            return None

        geojson = json.dumps(candidate.geometry)
        point = ogr.CreateGeometryFromJson(geojson)
        return point


class GeoContainsFilter(SpatialFilter):
    """
    指定した空間範囲に含まれていれば合格とするフィルタを作成します。

    Attributes
    ----------
    geo : osgeo.ogr.Geometry
        空間範囲を表す Geometry オブジェクト。 Polygon, MultiPolygon など、
        Contains() 演算が可能なタイプである必要があります。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.spatial_filter import GeoContainsFilter
    >>> api.init()
    >>> gcfilter = GeoContainsFilter({"type":"Polygon","coordinates":[[[139.6231,35.5461],[139.9106,35.5461],[139.9106,35.8271],[139.6231,35.8271],[139.6231,35.5461]]]})
    >>> # gcfilter はおおよそ23区を囲む矩形のフィルタ
    >>> lattice = api.analyze('新宿駅から八王子駅までは約40分です。')
    >>> for nodes in lattice:
    ...     [x.simple() for x in nodes]
    ...
    ["新宿駅(GEOWORD:['京王電鉄', '京王線'])", "新宿駅(GEOWORD:['小田急電鉄', '小田原線'])", "新宿駅(GEOWORD:['東京地下鉄', '4号線丸ノ内線'])", "新宿駅(GEOWORD:['東京都', '10号線新宿線'])", "新宿駅(GEOWORD:['東京都', '12号線大江戸線'])", "新宿駅(GEOWORD:['東日本旅客鉄道', '山手線'])", "新宿駅(GEOWORD:['東日本旅客鉄道', '中央線'])"]
    ['から(NORMAL)']
    ["八王子駅(GEOWORD:['東日本旅客鉄道', '横浜線'])", "八王子駅(GEOWORD:['東日本旅客鉄道', '中央線'])", "八王子駅(GEOWORD:['東日本旅客鉄道', '八高線'])"]
    ['まで(NORMAL)']
    ['は(NORMAL)']
    ['約(NORMAL)']
    ['40(NORMAL)']
    ['分(NORMAL)']
    ['です(NORMAL)']
    ['。(NORMAL)']
    >>> lattice_filtered = gcfilter(lattice)
    >>> # フィルタを適用すると「八王子駅」が範囲外なので地名語候補から除外されます
    >>> for nodes in lattice_filtered:
    ...     [x.simple() for x in nodes]
    ...
    ["新宿駅(GEOWORD:['京王電鉄', '京王線'])", "新宿駅(GEOWORD:['小田急電鉄', '小田原線'])", "新宿駅(GEOWORD:['東京地下鉄', '4号線丸ノ内線'])", "新宿駅(GEOWORD:['東京都', '10号線新宿線'])", "新宿駅(GEOWORD:['東京都', '12号線大江戸線'])", "新宿駅(GEOWORD:['東日本旅客鉄道', '山手線'])", "新宿駅(GEOWORD:['東日本旅客鉄道', '中央線'])"]
    ['から(NORMAL)']
    ['八王子駅(NORMAL)']
    ['まで(NORMAL)']
    ['は(NORMAL)']
    ['約(NORMAL)']
    ['40(NORMAL)']
    ['分(NORMAL)']
    ['です(NORMAL)']
    ['。(NORMAL)']

    """

    def __init__(self, geojson_or_url):
        super().__init__(geojson_or_url)

    def filter_func(self, candidate):
        point = self.__class__.point_from_candidate(candidate)
        if point is None:
            return True   # 座標を持たない候補は合格

        logger.debug("{} contains {} returns {}.".format(
            self.geo, point, self.geo.Contains(point)))
        return self.geo.Contains(point)


class GeoDisjointFilter(SpatialFilter):
    """
    指定した空間範囲に含まれていなければ合格とするフィルタを作成します。

    Attributes
    ----------
    geo : osgeo.ogr.Geometry
        空間範囲を表す Geometry オブジェクト。 Polygon, MultiPolygon など、
        Disjoint() 演算が可能なタイプである必要があります。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.spatial_filter import GeoDisjointFilter
    >>> api.init()
    >>> gdfilter = GeoDisjointFilter({"type":"Polygon","coordinates":[[[139.6231,35.5461],[139.9106,35.5461],[139.9106,35.8271],[139.6231,35.8271],[139.6231,35.5461]]]})
    >>> # gdfilter はおおよそ23区を囲む矩形のフィルタ
    >>> lattice = api.analyze('新宿駅から八王子駅までは約40分です。')
    >>> for nodes in lattice:
    ...     [x.simple() for x in nodes]
    ...
    ["新宿駅(GEOWORD:['京王電鉄', '京王線'])", "新宿駅(GEOWORD:['小田急電鉄', '小田原線'])", "新宿駅(GEOWORD:['東京地下鉄', '4号線丸ノ内線'])", "新宿駅(GEOWORD:['東京都', '10号線新宿線'])", "新宿駅(GEOWORD:['東京都', '12号線大江戸線'])", "新宿駅(GEOWORD:['東日本旅客鉄道', '山手線'])", "新宿駅(GEOWORD:['東日本旅客鉄道', '中央線'])"]
    ['から(NORMAL)']
    ["八王子駅(GEOWORD:['東日本旅客鉄道', '横浜線'])", "八王子駅(GEOWORD:['東日本旅客鉄道', '中央線'])", "八王子駅(GEOWORD:['東日本旅客鉄道', '八高線'])"]
    ['まで(NORMAL)']
    ['は(NORMAL)']
    ['約(NORMAL)']
    ['40(NORMAL)']
    ['分(NORMAL)']
    ['です(NORMAL)']
    ['。(NORMAL)']
    >>> lattice_filtered = gdfilter(lattice)
    >>> # フィルタを適用すると「新宿駅」が範囲内なので地名語候補から除外されます
    >>> for nodes in lattice_filtered:
    ...     [x.simple() for x in nodes]
    ...
    ['新宿駅(NORMAL)']
    ['から(NORMAL)']
    ["八王子駅(GEOWORD:['東日本旅客鉄道', '横浜線'])", "八王子駅(GEOWORD:['東日本旅客鉄道', '中央線'])", "八王子駅(GEOWORD:['東日本旅客鉄道', '八高線'])"]
    ['まで(NORMAL)']
    ['は(NORMAL)']
    ['約(NORMAL)']
    ['40(NORMAL)']
    ['分(NORMAL)']
    ['です(NORMAL)']
    ['。(NORMAL)']
    """

    def __init__(self, geojson_or_url):
        super().__init__(geojson_or_url)

    def filter_func(self, candidate):
        point = self.__class__.point_from_candidate(candidate)
        if point is None:
            return True  # 座標を持たない候補は合格

        return self.geo.Disjoint(point)



import logging


logger = logging.getLogger(__name__)


def pp_geojson(geojson_list, indent=2, file=None):
    """
    geoparse() の結果（GeoJSON互換）を pretty print します。

    Parameters
    ----------
    geojson_list : list
        geoparse() 結果の GeoJSON 互換 dict のリスト。
    indent : int, optional
        インデント幅。デフォルトは2です。
    file : file descriptor, optional
        出力先のファイルデスクリプタ。デフォルトは None です。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.devtool import pp_geojson
    >>> api.init()
    >>> pp_geojson(api.geoparse('アメリカ大使館：港区赤坂'))
    アメリカ大使館 ： 【港区赤坂:['東京都', '港区', '赤坂']】 EOS
    <BLANKLINE>
    """
    def simple(geojson):
        properties = geojson['properties']
        hypernym = ""

        if properties['node_type'] == "ADDRESS":
            return "【{}:{}】".format(
                properties['surface'],
                properties['address_properties']['fullname'])
        elif properties['node_type'] == "GEOWORD":
            hypernym = properties['geoword_properties'].get('hypernym')
            ne_class = properties['geoword_properties'].get('ne_class')
            if hypernym:
                return "≪{}|{}{}≫".format(
                    properties['surface'], ne_class, hypernym)

            return "≪{}|{}≫".format(properties['surface'], ne_class)

        return properties['surface']

    for pos, geojson in enumerate(geojson_list):
        print("{}".format(simple(geojson)), end=' ', file=file)

    print("EOS\n", file=file)


def pp_lattice(lattice, indent=2, file=None):
    """
    ラティス表現のデータを pretty print します。

    Parameters
    ----------
    lattice: list
        解析結果のラティス表現。
    indent: int, optional
        インデント幅。デフォルトは2です。
    file: file descriptor, optional
        出力先のファイルデスクリプタ。デフォルトは None です。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.devtool import pp_lattice
    >>> api.init()
    >>> parser = api.parser.Parser(jageocoder=True)
    >>> lattice = parser.analyze_sentence('アメリカ大使館：港区赤坂1-10-5')
    >>> pp_lattice(lattice)
    #0:'アメリカ大使館'
      アメリカ大使館(NORMAL)
    #1:'：'
      ：(NORMAL)
    #2:'港区'
      港区(GEOWORD:['東京都'])
      港区(GEOWORD:['愛知県', '名古屋市'])
      港区(GEOWORD:['大阪府', '大阪市'])
    #3:'赤坂'
      赤坂(GEOWORD:['上毛電気鉄道', '上毛線'])
      赤坂(GEOWORD:['東京地下鉄', '9号線千代田線'])
      赤坂(GEOWORD:['富士急行', '大月線'])
      赤坂(GEOWORD:['福岡市', '1号線(空港線)'])
    #4:'1'
      1(NORMAL)
    #5:'-'
      -(NORMAL)
    #6:'10'
      10(NORMAL)
    #7:'-'
      -(NORMAL)
    #8:'5'
      5(NORMAL)
    >>> lattice_address = parser.add_address_candidates(lattice, True)
    >>> pp_lattice(lattice_address)
    #0:'アメリカ大使館'
      アメリカ大使館(NORMAL)
    #1:'：'
      ：(NORMAL)
    #2:'港区'
      港区(GEOWORD:['東京都'])
      港区(GEOWORD:['愛知県', '名古屋市'])
      港区(GEOWORD:['大阪府', '大阪市'])
      港区赤坂1-10-(ADDRESS:東京都/港区/赤坂/一丁目/10番)[6]
    #3:'赤坂'
      赤坂(GEOWORD:['上毛電気鉄道', '上毛線'])
      赤坂(GEOWORD:['東京地下鉄', '9号線千代田線'])
      赤坂(GEOWORD:['富士急行', '大月線'])
      赤坂(GEOWORD:['福岡市', '1号線(空港線)'])
    #4:'1'
      1(NORMAL)
    #5:'-'
      -(NORMAL)
    #6:'10'
      10(NORMAL)
    #7:'-'
      -(NORMAL)
    #8:'5'
      5(NORMAL)
    >>> lattice_address_compact = parser.add_address_candidates(lattice)
    >>> pp_lattice(lattice_address_compact)
    #0:'アメリカ大使館'
      アメリカ大使館(NORMAL)
    #1:'：'
      ：(NORMAL)
    #2:'港区赤坂1-10-'
      港区赤坂1-10-(ADDRESS:東京都/港区/赤坂/一丁目/10番)[6]
    #3:'5'
      5(NORMAL)
    """
    for i in range(len(lattice)):
        nodes = lattice[i]
        print("#{}:'{}'".format(i, nodes[0].surface), file=file)
        for j, node in enumerate(nodes):
            print(" " * indent, end='', file=file)
            print(node.simple(), file=file)


def pp_path(path, indent=2, file=None):
    """
    パス表現のデータを pretty print します。

    Parameters
    ----------
    path: list
        解析結果のパス表現。
    indent: int, optional
        インデント幅。デフォルトは2です。
    file: file descriptor, optional
        出力先のファイルデスクリプタ。デフォルトは None です。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.linker import LinkedResults
    >>> from pygeonlp.api.devtool import pp_path
    >>> api.init(jageocoder=False)
    >>> lattice = api.analyze('アメリカ大使館：港区赤坂1-10-5')
    >>> for path in LinkedResults(lattice):
    ...   pp_path(path)
    ...
    [
      #0:アメリカ大使館(NORMAL)
      #1:：(NORMAL)
      #2:港区(GEOWORD:['東京都'])
      #3:赤坂(GEOWORD:['上毛電気鉄道', '上毛線'])
      #4:1(NORMAL)
      #5:-(NORMAL)
      #6:10(NORMAL)
      #7:-(NORMAL)
      #8:5(NORMAL)
    ]
    [
      #0:アメリカ大使館(NORMAL)
      #1:：(NORMAL)
      #2:港区(GEOWORD:['東京都'])
      #3:赤坂(GEOWORD:['東京地下鉄', '9号線千代田線'])
      #4:1(NORMAL)
      #5:-(NORMAL)
      #6:10(NORMAL)
      #7:-(NORMAL)
      #8:5(NORMAL)
    ]
    ...
    """
    print("[", file=file)
    for i in range(len(path)):
        node = path[i]
        print(" " * indent, end='', file=file)
        print("#{}:{}".format(i, node.simple()), file=file)

    print("]", file=file)


def _mecabline(geojson):
    """
    geoparse() の結果（GeoJSON表現）の一語分を mecab 類似書式用の
    リストに変換します。

    Parameters
    ----------
    geojson : dict
        geoparse() 結果の GeoJSON 互換 dict。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.devtool import _mecabline as mecabline
    >>> mecabline(api.geoparse('目黒駅')[0])
    ('目黒駅', ('名詞', '固有名詞', '地名語', '...:目黒駅', '*', '*', '目黒駅', '', ''), ('鉄道施設/鉄道駅', '...', '目黒駅', '139.71566', '35.632485'))
    """
    # Common properties
    properties = geojson["properties"]
    morphemes = properties["morphemes"]
    node_type = properties["node_type"]

    if node_type in ("NORMAL", "GEOWORD"):
        common = (
            morphemes["pos"],
            morphemes["subclass1"],
            morphemes["subclass2"],
            morphemes["subclass3"],
            morphemes["conjugation_type"],
            morphemes["conjugated_form"],
            morphemes["original_form"],
            morphemes["yomi"],
            morphemes["prononciation"],
        )

    if node_type == "NORMAL":
        geo = tuple()
    elif node_type == "GEOWORD":
        geoword_properties = geojson["properties"]["geoword_properties"]
        prefix = geoword_properties["prefix"][0] \
            if "prefix" in geoword_properties else ""
        suffix = geoword_properties["suffix"][0] \
            if "suffix" in geoword_properties else ""

        geo = (
            geoword_properties["ne_class"],
            geoword_properties["geolod_id"],
            prefix + geoword_properties["body"] + suffix,
            geoword_properties["longitude"],
            geoword_properties["latitude"],
        )

    else:  # ADDRESS
        address_properties = properties["address_properties"]
        levels = (
            "不明",
            "都道府県",
            "郡",
            "市町村・特別区",
            "区",
            "大字",
            "字・小字",
            "街区・地番",
            "住居番号・枝番"
        )
        level = levels[address_properties["level"]]
        original_form = "".join(address_properties["fullname"])
        yomi = "".join(m["properties"]["morphemes"]["yomi"]
                       for m in morphemes)
        prononciation = "".join(m["properties"]["morphemes"]
                                ["prononciation"] for m in morphemes)
        common = (
            "名詞",
            "固有名詞",
            "住所",
            level,
            "*",
            "*",
            original_form,
            yomi,
            prononciation,
        )
        geo = (
            "住所",
            "",
            original_form,
            str(address_properties["x"]),
            str(address_properties["y"]),
        )

    return (properties["surface"], common, geo,)


def pp_mecab(geojson_list, file=None):
    """
    geoparse() の結果（GeoJSON互換）を mecab 類似書式で出力します。

    Parameters
    ----------
    geojson_list : list
        geoparse() 結果の GeoJSON 互換 dict のリスト。
    file : file descriptor, optional
        出力先のファイルデスクリプタ。デフォルトは None です。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.devtool import pp_mecab
    >>> pp_mecab(api.geoparse('目黒駅は品川区にあります。'))
    目黒駅  名詞,固有名詞,地名語,Xy26iV:目黒駅,*,*,目黒駅,, 鉄道施設/鉄道駅,Xy26iV,目黒駅,139.71566,35.632485
    は      助詞,係助詞,*,*,*,*,は,ハ,ワ
    品川区  名詞,固有名詞,地名語,kEAYBl:品川区,*,*,品川区,, 市区町村,kEAYBl,品川区,139.73025000,35.60906600
    に      助詞,格助詞,一般,*,*,*,に,ニ,ニ
    あり    動詞,自立,*,*,連用形,五段・ラ行,ある,アリ,アリ
    ます    助動詞,*,*,*,基本形,特殊・マス,ます,マス,マス
    。      記号,句点,*,*,*,*,。,。,。
    EOS
    <BLANKLINE>
    """  # noqa: E501

    for geojson in geojson_list:
        if geojson is not None:
            line = _mecabline(geojson)
            print("\t".join((
                line[0],
                ",".join(line[1]),
                ",".join(line[2]),
            )), end="\n", file=file
            )

    print("EOS", file=file)



import json
from logging import getLogger
import math

from geographiclib.geodesic import Geodesic

logger = getLogger(__name__)

try:
    from osgeo import ogr
    have_gdal = True
except ModuleNotFoundError:
    logger.info(("gdal パッケージがインストールされていないため、"
                 "ノード間の地理的距離計算をスキップします。"))
    have_gdal = False


class Node(object):
    """
    形態素ノードを表すクラス。

    service.analyze() が返すラティス表現や、 linker.Evaluator.get() が
    返すパス表現では、形態素は Node インスタンスとして格納されます。

    Attributes
    ----------
    surface : str
        表記文字列。
    node_type : int
        ノードの種類。地名語ノードは1、住所ノードは2、それ以外は0です。
    morphemes : list of dict
        形態素の属性。
    geometry : object or None
        地理属性。
    prop : dict
        その他の属性（geowordまたは住所）。
    """
    IGNORE = -1
    NORMAL = 0
    GEOWORD = 1
    ADDRESS = 2

    def __init__(self, surface, node_type, morphemes,
                 geometry=None, prop=None):
        """
        ノード情報を初期化します。

        ノードは基本的に Parser が作成したものを利用してください。
        """

        self.surface = surface
        self.node_type = node_type
        self.morphemes = morphemes
        self.geometry = geometry
        self.prop = prop
        self._attr = {}  # 動的に計算する属性

    def get_lonlat(self):
        """
        ノードの経度、緯度を返します。

        Returns
        -------
        dict
            以下の要素を持つ dict。

            - lat : float
                ノードの代表点の緯度。
            - lon : float
                ノードの代表点の経度。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> api.init()
        >>> node = api.analyze('国会議事堂前')[0][0]
        >>> node.get_lonlat()
        {'lat': 35.674845, 'lon': 139.74534166666666}

        >>> import pygeonlp.api as api
        >>> api.init()
        >>> node = api.analyze('千代田区一ツ橋2-1-2', jageocoder=True)[0][0]
        >>> node.get_lonlat()
        {'lat': 35.69..., 'lon': 139.75...}
        """
        if self.node_type == self.__class__.GEOWORD:
            return {
                'lat': float(self.prop['latitude']),
                'lon': float(self.prop['longitude']),
            }

        if self.node_type == self.__class__.ADDRESS:
            return {
                'lat': self.prop['y'],
                'lon': self.prop['x'],
            }

        return None

    def __get_type_string(self):
        if self.node_type == self.__class__.NORMAL:
            return "NORMAL"
        elif self.node_type == self.__class__.GEOWORD:
            return "GEOWORD"
        elif self.node_type == self.__class__.ADDRESS:
            return "ADDRESS"

        return "UNKNOWN"

    def __repr__(self):
        return json.dumps(self.as_dict(), ensure_ascii=False)

    def simple(self):
        """
        ノード情報をシンプルな形式で表現します。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> api.init()
        >>> api.analyze('国会議事堂前')[0][0].simple()
        "国会議事堂前(GEOWORD:['東京地下鉄', '4号線丸ノ内線'])"

        Returns
        -------
        str
            シンプルな形式の文字列。
        """

        hypernym = None
        fullname = None
        if self.node_type == Node.GEOWORD:
            hypernym = self.prop.get('hypernym')
            fullname = self.prop.get('fullname')

        if self.node_type == Node.ADDRESS:
            res = "{}({}:{})[{}]".format(
                self.surface,
                self.__get_type_string(),
                "/".join(self.prop['fullname']),
                len(self.morphemes))
        elif hypernym:
            res = "{}({}:{})".format(self.surface,
                                     self.__get_type_string(), hypernym)
        elif fullname:
            res = "{}({},{})".format(
                self.surface, self.__get_type_string(), fullname)
        else:
            res = "{}({})".format(
                self.surface, self.__get_type_string())

        return res

    def as_dict(self):
        """
        ノード情報を JSON 出力可能な dict オブジェクトに変換します。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> api.init()
        >>> api.analyze('国会議事堂前')[0][0].as_dict()
        {'surface': '国会議事堂前', 'node_type': 'GEOWORD', 'morphemes': {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '国会議事堂前', 'pos': '名詞', 'prononciation': '', 'subclass1': '固有名詞', 'subclass2': '地名語', 'subclass3': 'Bn4q6d:国会議事堂前駅', 'surface': '国会議事堂前', 'yomi': ''}, 'geometry': {'type': 'Point', 'coordinates': [139.74534166666666, 35.674845]}, 'prop': {'body': '国会議事堂前', 'dictionary_id': 3, 'entry_id': 'LrGGxY', 'geolod_id': 'Bn4q6d', 'hypernym': ['東京地下鉄', '4号線丸ノ内線'], 'institution_type': '民営鉄道', 'latitude': '35.674845', 'longitude': '139.74534166666666', 'ne_class': '鉄道施設/鉄道駅', 'railway_class': '普通鉄道', 'suffix': ['駅', ''], 'dictionary_identifier': 'geonlp:ksj-station-N02'}}

        Returns
        -------
        dict
            JSON 出力可能な dict オブジェクト。
        """  # noqa: E501
        if isinstance(self.morphemes, list):
            morphemes = [x.as_dict() for x in self.morphemes]
        else:
            morphemes = self.morphemes

        obj = {
            "surface": self.surface,
            "node_type": self.__get_type_string(),
            "morphemes": morphemes,
            "geometry": self.geometry,
            "prop": self.prop,
        }
        return obj

    def as_geojson(self):
        """
        ノード情報を GeoJSON の Feature 形式にダンプ可能な dict オブジェクトに変換します。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> api.init()
        >>> api.analyze('国会議事堂前')[0][0].as_geojson()
        {'type': 'Feature', 'geometry': {'type': 'Point', 'coordinates': [139.74534166666666, 35.674845]}, 'properties': {'surface': '国会議事堂前', 'node_type': 'GEOWORD', 'morphemes': {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '国会議事堂前', 'pos': '名詞', 'prononciation': '', 'subclass1': '固有名詞', 'subclass2': '地名語', 'subclass3': 'Bn4q6d:国会議事堂前駅', 'surface': '国会議事堂前', 'yomi': ''}, 'geoword_properties': {'body': '国会議事堂前', 'dictionary_id': 3, 'entry_id': 'LrGGxY', 'geolod_id': 'Bn4q6d', 'hypernym': ['東京地下鉄', '4号線丸ノ内線'], 'institution_type': '民営鉄道', 'latitude': '35.674845', 'longitude': '139.74534166666666', 'ne_class': '鉄道施設/鉄道駅', 'railway_class': '普通鉄道', 'suffix': ['駅', ''], 'dictionary_identifier': 'geonlp:ksj-station-N02'}}}

        Returns
        -------
        dict
            GeoJSON Feature 形式に変換可能な dict オブジェクト。
        """  # noqa: E501
        if isinstance(self.morphemes, list):
            # 住所の場合、ネストしている内部の Node も GeoJSON dict に展開する
            morphemes = [x.as_geojson() for x in self.morphemes]
        else:
            morphemes = self.morphemes

        feature = {
            "type": "Feature",
            "geometry": self.geometry,
            "properties": {
                "surface": self.surface,
                "node_type": self.__get_type_string(),
                "morphemes": morphemes,
            }
        }

        if self.node_type == Node.GEOWORD:
            feature['properties']['geoword_properties'] = self.prop
        elif self.node_type == Node.ADDRESS:
            feature['properties']['address_properties'] = self.prop

        return feature

    def _add_node(self, surface, node):
        """
        形態素ノードをブロックに追加します。

        Parameters
        ----------
        surface : str
            ノードの表記
        node : dict
            形態素ノード、地名語ノードまたは住所ノード
        """
        self.surface += surface
        self.nodes.append(node)

    def get_notations(self):
        """
        このノードの表記として可能性があるものの候補を取得し、
        set を返します（重複を除去し、 AND を高速に計算するため）。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> api.init()
        >>> # set は順番を保持しないので sorted で昇順にソート
        >>> sorted(api.analyze('国会議事堂前')[0][0].get_notations())
        ['国会議事堂前', '国会議事堂前駅']

        Note
        ----
        何度も計算しないように、計算した結果は ``_attr['notations']`` に保持します。
        """
        if 'notations' in self._attr:
            return self._attr['notations']

        if self.node_type != Node.GEOWORD:
            self._attr['notations'] = set(self.surface,)
            return self._attr['notations']

        body = self.prop.get('body')
        prefixes = self.prop.get('prefix', None)
        suffixes = self.prop.get('suffix', None)
        cands = [body]
        if prefixes:
            for prefix in prefixes:
                cands.append(prefix + body)

        if suffixes:
            for suffix in suffixes:
                cands.append(body + suffix)

        if prefixes and suffixes:
            for prefix in prefixes:
                for suffix in suffixes:
                    cands.append(prefix + body + suffix)

        self._attr['notations'] = set(cands)
        return self._attr['notations']

    def get_point_object(self):
        """
        地名語ノードまたは住所ノードの場合、経度緯度から Point オブジェクトを作成します。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> api.init()
        >>> api.analyze('国会議事堂前')[0][0].get_point_object().ExportToWkt()
        'POINT (139.745341666667 35.674845)'

        Note
        ----
        ``gdal`` がインストールされていない場合は None を返します。
        何度も計算しないように、計算した結果は ``_attr['point']`` に保持します。
        """
        if 'point' in self._attr:
            return self._attr['point']

        if have_gdal is False or self.geometry is None:
            return None

        point = ogr.CreateGeometryFromJson(json.dumps(self.geometry))
        self._attr['point'] = point
        return point

    def distance(self, node):
        """
        自身と他のノードの距離 (m) を計算します。
        このメソッドは geographiclib を利用します。
        https://pypi.org/project/geographiclib/

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> api.init()
        >>> lattice = api.analyze('国会議事堂前,永田町')
        >>> n0 = lattice[0][0]
        >>> n1 = lattice[2][0]
        >>> round(n0.distance(n1), 6)
        676.592019
        """
        p0 = self.get_lonlat()
        p1 = node.get_lonlat()

        if p0 is None or p1 is None:
            raise RuntimeError(
                "ジオメトリを持たないノード間の距離は計算できません。")

        geod = Geodesic.WGS84
        g = geod.Inverse(p0['lat'], p0['lon'], p1['lat'], p1['lon'])
        dist = g['s12']

        if math.isnan(dist):
            raise RuntimeError(
                "距離が計算できません（経緯度が不正な値）。")

        return dist



import logging
import os

import jageocoder
import pygeonlp.api as geonlp_api


class BaseConfig(object):
    DEBUG = False
    TESTING = False

    GEONLP_DIR = geonlp_api.get_db_dir()
    MECAB_DIC_DIR = os.environ.get('GEONLP_MECAB_DIC_DIR')
    GEONLP_API_OPTIONS = {}

    if MECAB_DIC_DIR is not None:
        GEONLP_API_OPTIONS['system_dic_dir'] = MECAB_DIC_DIR

    try:
        jageocoder.init()
        JAGEOCODER_TREE = jageocoder.get_module_tree()
    except jageocoder.exceptions.JageocoderError:
        JAGEOCODER_TREE = None

    LOGGING = {
        'version': 1,
        'formatters': {'default': {
            'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        }},
        'handlers': {'wsgi': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://flask.logging.wsgi_errors_stream',
            'formatter': 'default'
        }},
        'root': {
            'level': logging.INFO,
            'handlers': ['wsgi']
        }
    }


class ProductionConfig(BaseConfig):
    DEBUG = False


class DevelopmentConfig(BaseConfig):
    DEBUG = True


class TestingConfig(BaseConfig):
    TESTING = True


config = DevelopmentConfig()



from io import StringIO
import json

import chardet
from lxml import etree
import requests


class MetadataError(RuntimeError):
    """
    メタデータの操作の際に例外が起こると、このクラスが発生します。
    """
    pass


class Metadata(object):
    """
    JSON-LD 形式メタデータの管理クラス。

    Attributes
    ----------
    jsonld : str
        メタデータ json-ld 文字列。
    """

    def __init__(self, jsonld=None):
        """
        Parameters
        ----------
        jsonld : str or dict
            json-ld 文字列、またはデコードした dict。
        """
        if isinstance(jsonld, dict):
            jsonld = json.dumps(jsonld, ensure_ascii=False)

        self.jsonld = jsonld

    def __repr__(self):
        return self.jsonld

    @classmethod
    def download(cls, url, params=None, **kwargs):
        """
        指定した URL からウェブページをダウンロードし、
        ヘッダに記載されている json-ld を辞書メタデータとして抽出します。

        Parameters
        ----------
        url : str
            json-ld を含むウェブページの URL。
        params : dict, optional
            requests.get に渡す params パラメータ。
        **kwargs : キーワードパラメータ, optional
            requests.get に渡す kwargs パラメータ。

        Returns
        -------
        Metadata
            抽出した json-ld 文字列をセットした Metadata インスタンス

        Raises
        ----------
        MetadataError
            ウェブページのダウンロードに失敗した場合。
            ウェブページから json-ld が見つけられない場合。
            ウェブページに複数の json-ld が含まれている場合。
        """
        r = requests.get(url, params, **kwargs)
        if r.status_code != 200:
            raise MetadataError(
                "指定されたページを取得できません。status code={}".format(
                    r.status_code))

        htmldata = r.content
        encoding = chardet.detect(htmldata)
        htmltext = htmldata.decode(encoding['encoding'])

        tree = etree.parse(StringIO(htmltext), etree.HTMLParser())
        p = tree.findall(".//script[@type='application/ld+json']")
        if len(p) == 0:
            raise MetadataError(
                "指定されたページから json-ldを見つけられません。")

        if len(p) > 1:
            raise MetadataError(
                "指定されたページには複数の json-ldが含まれています。")

        jsonld = p[0].text
        return cls(jsonld)

    @classmethod
    def load(cls, file):
        """
        指定したファイルに含まれる json-ld 文字列から
        Metadata インスタンスを作成します。

        Parameters
        ----------
        file : str
            json-ld 文字列を含むファイル名。

        Returns
        -------
        Metadata
            json-ld 文字列をセットした Metadata インスタンス。
        """
        with open(mode='r', file=file, encoding='utf-8') as f:
            jsonld = f.read()

        return cls(jsonld)

    @classmethod
    def loads(cls, jsonld):
        """
        指定した json-ld 文字列から Metadata インスタンスを作成します。

        Parameters
        ----------
        jsonld : str
            json-ld 文字列。

        Returns
        -------
        Metadata
            json-ld 文字列をセットした Metadata インスタンス。
        """
        return cls(jsonld)

    def get_name(self):
        """
        json-ld 文字列からデータセットの名前を抽出します。

        Returns
        -------
        str
            データセットの名前。

        Raises
        ----------
        MetadataError
            json-ld がセットされていない場合。
            json-ld から identifier が見つけられない場合。
        """
        if self.jsonld is None:
            raise MetadataError(
                "json-ld がセットされていません。")

        obj = json.loads(self.jsonld)
        try:
            name = obj['name']
            return name
        except (IndexError, TypeError):
            raise MetadataError(
                ("json-ld から name を見つけられません。"
                 "{}".format(self.jsonld)))

        raise MetadataError(
            ("json-ldから name を見つけられません。"
             "{}".format(self.jsonld)))
        return None

    def get_identifier(self):
        """
        json-ld 文字列から GeoNLP identifier を抽出します。

        GeoNLP identifier は 'geonlp:' で始まる必要があります。

        Returns
        -------
        str
            抽出した identifier 文字列。

        Raises
        ----------
        MetadataError
            json-ld がセットされていない場合。
            json-ld から identifier が見つけられない場合。
        """
        if self.jsonld is None:
            raise MetadataError(
                "json-ld がセットされていません。")

        obj = json.loads(self.jsonld)
        try:
            identifiers = obj['identifier']
            if isinstance(identifiers, str):
                identifiers = [identifiers]

            for identifier in identifiers:
                if identifier.startswith('geonlp:'):
                    return identifier

        except (IndexError, TypeError):
            raise MetadataError(
                ("json-ldから 'geonlp:' で始まる identifier を見つけられません。"
                 "{}".format(self.jsonld)))

        raise MetadataError(
            ("json-ldから identifier を見つけられません。"
             "{}".format(self.jsonld)))
        return None

    def get_content_url(self):
        """
        json-ld 文字列から、コンテンツ CSV をダウンロードする
        content_url を抽出します。



        Returns
        -------
        str
            抽出した content_url 文字列。

        Raises
        ----------
        MetadataError
            json-ld がセットされていない場合。
            json-ld から content_url が見つけられない場合。
        """
        content_url = None
        if self.jsonld is None:
            raise MetadataError(
                "json-ld がセットされていません。")

        obj = json.loads(self.jsonld)
        try:
            for dist in obj['distribution']:
                if '@type' not in dist or dist['@type'] != "DataDownload":
                    continue

                content_url = dist['contentUrl']
        except (IndexError, TypeError):
            raise MetadataError(
                ("json-ld から contentUrl を見つけられません。"
                 "{}".format(self.jsonld)))

        if content_url is None:
            raise MetadataError(
                ("json-ld から contentUrl を見つけられません。"
                    "{}".format(self.jsonld)))

        self.content_url = content_url
        return content_url

    def download_csv(self, content_url=None, params=None, **kwargs):
        """
        content_url から CSV をダウンロードします。

        Parameters
        ----------
        content_url : str, optional
            CSV をダウンロードする URL を指定します。
            省略された場合、 json-ld から content_url を抽出します。
        params : dict, optional
            requests.get に渡す params パラメータ。
        **kwargs : キーワードパラメータ, optional
            requests.get に渡す kwargs パラメータ。

        Returns
        -------
        str
            ダウンロードした CSV テキスト。

        Raises
        ----------
        MetadataError
            CSV のダウンロードに失敗した場合。
        """
        if content_url is None:
            content_url = self.get_content_url()

        r = requests.get(content_url, params, **kwargs)
        if r.status_code != 200:
            raise MetadataError(
                "CSV をダウンロードできません。 url={}".format(content_url))

        csvdata = r.content
        encoding = chardet.detect(csvdata)
        csvtext = csvdata.decode(encoding['encoding'])

        return csvtext



import logging

from pygeonlp.api.parser import Parser
from pygeonlp.api.filter import InputBasedFilter
from pygeonlp.api.linker import Evaluator
from pygeonlp.api.node import Node

logger = logging.getLogger(__name__)


class WorkflowError(RuntimeError):
    """
    ワークフロー処理の際に例外が起こると、このクラスが発生します。
    """
    pass


class Workflow(object):
    """
    一連のジオパース処理を行なうワークフロー。

    Attributes
    ----------
    parser : pygeonlp.parser.Parser
        テキストを解析し、地名語を埋め込む Parser のインスタンス。
        Parser はテキスト（文字列）からラティス表現を作成します。
    filters : list of pygeonlp.filter.Filter
        ラティス表現に含まれる候補に対し、条件による選別を行い、
        候補を絞り込む Filter インスタンスのリスト。
        リストの順番に Filter を適用します。
    evaluator : pygeonlp.linker.Evaluator
        ラティス表現から順位付きパス表現を作成する Evaluator のインスタンス。
    """

    def __init__(
            self,
            db_dir=None,
            address_regex=None,
            jageocoder=None,
            scoring_class=None,
            scoring_options=None,
            filters=None,
            max_combinations=None,
            **options
        ):
        """
        パーザを初期化します。

        Parameters
        ----------
        db_dir : PathLike, optional
            データベースディレクトリ。
            省略した場合は ``api.init.get_db_dir()`` が返す値を利用します。
        address_regex : str, optional
            住所表記の開始とみなす地名語の固有名クラスを表す正規表現。
            省略した場合、``r'^(都道府県|市区町村|行政地域|居住地名)(/.+|)'``
            を利用します。
        jageocoder : jageocoder.tree.AddressTree, optional
            利用する住所ジオコーダーを指定します。省略した場合、
            jageocoder モジュールのデフォルトオブジェクトを利用します。
            False を指定した場合、ジオコーディング機能を利用しません。
        scoring_class : class, optional
            パスのスコアとノード間のスコアを計算する関数を持つ
            スコアリングクラス。
            指定しない場合、``pygeonlp.api.scoring`` モジュール内の
            ``ScoringClass`` が利用されます。
        scoring_options : any, optional
            スコアリングクラスの初期化に渡すオプションパラメータ。
        filters: list
            適用するフィルタオブジェクトのリスト。
            省略した場合は InputBasedFilter がセットされます。
        max_combinations: int, optional
            ノード候補の組み合わせ数の上限値。これを超える組み合わせが
            可能な入力が与えられた場合は例外 LinkerError を発生しますので
            必要に応じて大きくしてください。
            デフォルト値は linker.MAX_COMBINATIONS です。
        """
        self.parser = Parser(
            db_dir=db_dir,
            address_regex=address_regex,
            jageocoder=jageocoder,
            **options)
        self.filters = filters or [InputBasedFilter()]
        self.evaluator = Evaluator(
            scoring_class=scoring_class,
            scoring_options=scoring_options,
            max_results=1,
            max_combinations=max_combinations)

    def geoparse(self, sentence: str):
        """
        文を解析して GeoJSON Feature 形式に変換可能な dict の
        リストを返します。

        Parameters
        ----------
        sentence: str
            解析する文字列

        Returns
        -------
        list
            GeoJSON Feature 形式に変換可能な dict のリスト。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> api.init()
        >>> workflow = api.workflow.Workflow()
        >>> workflow.geoparse('国会議事堂前まで歩きました。')
        [{'type': 'Feature', 'geometry': {'type': 'Point', 'coordinates': [139.74305333333334, 35.673543333333335]}, 'properties': {'surface': '国会議事堂前', 'node_type': 'GEOWORD', 'morphemes': {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '国会議事堂前', 'pos': '名詞', 'prononciation': '', 'subclass1': '固有名詞', 'subclass2': '地名語', 'subclass3': 'cE8W4w:国会議事堂前駅', 'surface': '国会議事堂前', 'yomi': ''}, 'geoword_properties': {'body': '国会議事堂前', 'dictionary_id': 3, 'entry_id': '4NFELa', 'geolod_id': 'cE8W4w', 'hypernym': ['東京地下鉄', '9号線千代田線'], 'institution_type': '民営鉄道', 'latitude': '35.673543333333335', 'longitude': '139.74305333333334', 'ne_class': '鉄道施設/鉄道駅', 'railway_class': '普通鉄道', 'suffix': ['駅', ''], 'dictionary_identifier': 'geonlp:ksj-station-N02'}}}, {'type': 'Feature', 'geometry': None, 'properties': {'surface': 'まで', 'node_type': 'NORMAL', 'morphemes': {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': 'まで', 'pos': '助詞', 'prononciation': 'マデ', 'subclass1': '副助詞', 'subclass2': '*', 'subclass3': '*', 'surface': 'まで', 'yomi': 'マデ'}}}, {'type': 'Feature', 'geometry': None, 'properties': {'surface': '歩き', 'node_type': 'NORMAL', 'morphemes': {'conjugated_form': '五段・カ行イ音便', 'conjugation_type': '連用形', 'original_form': '歩く', 'pos': '動詞', 'prononciation': 'アルキ', 'subclass1': '自立', 'subclass2': '*', 'subclass3': '*', 'surface': '歩き', 'yomi': 'アルキ'}}}, {'type': 'Feature', 'geometry': None, 'properties': {'surface': 'まし', 'node_type': 'NORMAL', 'morphemes': {'conjugated_form': '特殊・マス', 'conjugation_type': '連用形', 'original_form': 'ます', 'pos': '助動詞', 'prononciation': 'マシ', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': 'まし', 'yomi': 'マシ'}}}, {'type': 'Feature', 'geometry': None, 'properties': {'surface': 'た', 'node_type': 'NORMAL', 'morphemes': {'conjugated_form': '特殊・タ', 'conjugation_type': '基本形', 'original_form': 'た', 'pos': '助動詞', 'prononciation': 'タ', 'subclass1': '*', 'subclass2': '*', 'subclass3': '*', 'surface': 'た', 'yomi': 'タ'}}}, {'type': 'Feature', 'geometry': None, 'properties': {'surface': '。', 'node_type': 'NORMAL', 'morphemes': {'conjugated_form': '*', 'conjugation_type': '*', 'original_form': '。', 'pos': '記号', 'prononciation': '。', 'subclass1': '句点', 'subclass2': '*', 'subclass3': '*', 'surface': '。', 'yomi': '。'}}}]
        """
        lattice = self.parser.analyze_sentence(sentence)

        # フィルタを適用
        for f in self.filters:
            lattice = f(lattice)

        # 住所候補を追加
        lattice = self.parser.add_address_candidates(lattice)

        # 処理可能な長さに分割したラティスをパス表現に変換
        results = []
        for lattice_part in self.get_processible_lattice_part(lattice):
            if len(lattice_part) < 1:
                continue    # 空行

            results += self.evaluator.get(lattice_part)

        features = []
        for result in results:
            nodes = result['result']
            for node in nodes:
                features.append(node.as_geojson())

        return features

    def get_processible_lattice_part(self, lattice):
        """
        組み合わせの候補数が MAX_COMBINATIONS 未満になるように
        ラティスの先頭部分から区切りの良い一部分を抽出するジェネレータ。

        Parameters
        ----------
        lattice: list
            入力となるラティス表現。

        Return
        ------
        list
            先頭部分から切り出した部分的なラティス表現。

        Note
        ----
        この関数はジェネレータなので yield で返します。

        入力テキストの区切り文字が事前に分かっている場合や、
        途中で分割されたくない場合には、意図した位置で分割するように
        オーバーライドしてください。
        """
        pos_from = 0
        pos_to = len(lattice)

        while pos_from < len(lattice):
            lattice_part = lattice[pos_from:pos_to]
            if pos_to - pos_from == 1 or \
                    self.evaluator.count_combinations(lattice_part) \
                        < self.evaluator.max_combinations:
                logger.debug("--- pos {} - {}".format(pos_from, pos_to))
                for i in range(pos_from, pos_to):
                    nodes = lattice[i]
                    logger.debug("{}:'{}' has {} nodes".format(
                        i, nodes[0].surface, len(nodes)))
                logger.debug("---")
                yield lattice_part

                pos_from = pos_to
                pos_to = len(lattice)
                continue

            # 組み合わせ数が多すぎるので分割する
            eliminated = False

            # 句点で分割してみる
            for i in range(pos_from, pos_to - 1):
                node = lattice[i][0]  # i 番目のノード集合の先頭
                if node.node_type != Node.ADDRESS and \
                        node.morphemes['subclass1'] == '句点':
                    pos_to = i + 1
                    eliminated = True
                    break

            if eliminated:
                continue

            # 改行で分割してみる
            for i in range(pos_from, pos_to - 1):
                node = lattice[i][0]  # i 番目のノード集合の先頭
                if node.node_type != Node.ADDRESS and \
                    self.parser.check_word(
                        node.morphemes,
                        {'pos': '記号',
                         'subclass1': '制御コード',
                                      'subclass2': '改行'}):
                    pos_to = i + 1
                    eliminated = True
                    break

            if eliminated:
                continue

            # 記号で分割してみる
            for i in range(pos_from, pos_to - 1):
                node = lattice[i][0]  # i 番目のノード集合の先頭
                if node.node_type != Node.ADDRESS and \
                    self.parser.check_word(node.morphemes,
                                           {'pos': '記号',
                                            'subclass1': '一般'}) and \
                        node.surface in ('／/★●○◎■□◇'):
                    pos_to = i + 1
                    eliminated = True
                    break

            if eliminated:
                continue

            # 読点で分割してみる
            for i in range(pos_from, pos_to - 1):
                node = lattice[i][0]  # i 番目のノード集合の先頭
                if node.node_type != Node.ADDRESS and \
                        node.morphemes['subclass1'] == '読点':
                    pos_to = i + 1
                    eliminated = True
                    break

            if eliminated:
                continue

            # 半分にする、ただし住所表現は分割しない
            i = pos_from
            while i < pos_to:
                logger.debug(
                    '半分 i={i}, pos_from={pos_from}, pos_to={pos_to}'.format(
                        i=i, pos_from=pos_from, pos_to=pos_to))
                if i >= pos_from + int((pos_to - pos_from) / 2):
                    pos_to = i
                    break

                has_non_address_node = False
                for node in lattice[i]:
                    if node.node_type != Node.ADDRESS:
                        has_non_address_node = True
                        break

                node_width = 1
                if has_non_address_node:
                    # 住所ノード以外も保存されている場合
                    for node in lattice[i]:
                        if isinstance(node.morphemes, list):
                            # 住所ノードは複数の形態素を含む
                            node_width = len(node.morphemes)
                            break

                i += node_width

    def getActiveDictionaries(self):
        return self.parser.service.getActiveDictionaries()

    def setActiveDictionaries(self, idlist=None, pattern=None):
        return self.parser.service.setActiveDictionaries(idlist, pattern)

    def disactivateDictionaries(self, idlist=None, pattern=None):
        return self.parser.service.disactivateDictionaries(idlist, pattern)

    def activateDictionaries(self, idlist=None, pattern=None):
        return self.parser.service.activateDictionaries(idlist, pattern)

    def getActiveClasses(self):
        return self.parser.service.getActiveClasses()

    def setActiveClasses(self, patterns=None):
        return self.parser.service.setActiveClasses(patterns)



import copy
import logging
import re

from pygeonlp.api.node import Node

logger = logging.getLogger(__name__)


class FilterError(RuntimeError):
    """
    フィルタ処理の際に例外が起こると、このクラスが発生します。
    """
    pass


class Filter(object):
    """
    ラティス表現を入力として、条件に一致しないノードを除外した結果を
    ラティス表現として出力するフィルタ機能の基底クラス。

    このクラスは全てのフィルタに共通の機能を提供するものなので、
    派生クラスを利用してください。
    """

    def __init__(self, **kwargs):
        """
        フィルタを初期化します。
        初期化の内容は派生するクラスごとに拡張する必要があります。

        Attributes
        ----------
        when_all_failed : str
            全ての候補が条件を満たさなかった場合の挙動を指定します。

            "return_all"
                全ての候補をそのまま残します。
            "convert_to_normal"
                '名詞・固有名詞・地域・一般' ノードを1つ作成して返します。
        """
        self.when_all_failed = 'return_all'

    def apply(self, input, **kwargs):
        """
        入力データの各形態素に対して ``apply_filter()`` を適用し、
        合格したノードだけを残した結果を返します。

        ``apply_filter()`` では一つの形態素に対応する候補ノードの
        リストに含まれる情報のみを利用してフィルタ処理を行なうため、
        前後の形態素を見て判定する必要があるフィルタクラスでは、
        ``apply()`` メソッドを上書きする必要があります。

        Parameters
        ----------
        input : list
            入力となるラティス表現。

        Returns
        -------
        list
            ``filter_func()`` を適用して要素を削除されたラティス表現。
        """
        output = [None] * len(input)
        for i in range(len(input)):
            candidates = input[i]
            output[i] = self.apply_filter(candidates, **kwargs)

            if len(output[i]) > 0:
                continue

            # 1つも候補が残らなかった場合
            if self.when_all_failed == 'return_all':
                # 入力の候補リストをそのまま残す
                output[i] = input[i]
                continue

            if self.when_all_failed == 'convert_to_normal':
                # 非地名語ノードがあればそれを採用する
                for node in candidates:
                    if node.node_type == Node.NORMAL:
                        output[i] = [node]
                        break

                if len(output[i]) > 0:
                    continue

                # 名詞・固有名詞・地域・一般ノードを作成して候補とする
                new_node = copy.copy(candidates[0])
                new_node.node_type = Node.NORMAL
                new_node.morphemes.update({
                    "conjugated_form": "*",
                    "conjugation_type": "*",
                    "pos": "名詞",
                    "subclass1": "固有名詞",
                    "subclass2": "地域",
                    "subclass3": "一般",
                })
                new_node.geometry = None
                new_node.prop = None
                output[i] = [new_node]
                continue

            raise FilterError(
                "定義されていない動作:'{}'".format(self.when_all_failed))

        return output

    def apply_filter(self, candidates, default=None, **kwargs):
        '''
        ラティス表現の1つの形態素に対する複数の候補ノードに対して
        順番に ``filter_func()`` を適用し、条件に一致しないノードを
        削除した候補ノードのリストを返します。

        default で指定された地名語 ID を持つノードが条件に一致した場合、
        結果候補ノードリストの先頭に配置します。

        ``filter_func()`` では個々の候補ノードに含まれる情報のみを
        利用してフィルタへの合否判定処理を行なうため、
        1つの形態素に対する候補ノードを比較して最適なものを
        選択する必要があるフィルタクラスでは、
        ``apply_filter()`` メソッドを上書きする必要があります。

        Parameters
        ----------
        candidates : list
            候補ノードのリスト
        default : str, optional
            最優先候補としたいノードの geolod_id
        **kwargs
            キーワードパラメータ、派生クラスでは利用する可能性があります。

        Returns
        -------
        list
            条件に一致した候補ノードのリスト。
        '''
        top_candidate = None
        passed = []
        excluded = []

        for candidate in candidates:
            if self.filter_func(candidate):
                if candidate.prop is None:
                    # 地名語・住所以外の形態素
                    passed.append(candidate)
                    continue

                if top_candidate is None and \
                   'geolod_id' in candidate.prop and \
                   default == candidate.prop['geolod_id']:
                    top_candidate = candidate
                else:
                    passed.append(candidate)
            else:
                excluded.append(candidate)

        if top_candidate:
            passed.insert(0, top_candidate)

        return passed

    def filter_func(candidate):
        """
        候補ノードが条件に一致するかどうかを判定します。
        派生クラスでそれぞれ実装する必要があります。

        Parameters
        ----------
        candidate : Node
            判定する形態素ノード。

        Returns
        -------
        bool
            基底クラスでは常に True を返します。
        """
        return True

    def __call__(self, input):
        """
        クラスをメソッドのように呼びだせるようにするためのマジックメソッド。
        Filter(input) は Filter.apply(input) を実行します。
        """
        return self.apply(input)

    @staticmethod
    def count_geowords(lattice):
        """
        ラティス表現を高速に粗く解析し、
        地名語候補を含む形態素の数や ne_class の分布などを集計します。

        詳細な解析を行うためのプランを決定する前処理として利用します。

        Parameters
        ----------
        lattice: list
            ラティス表現。

        Returns
        -------
        dict
            以下の要素を含む集計結果を返します。

            num_geowords: int
                地名語候補を1つ以上含む形態素ノードの数。
            num_addresses: int
                住所候補を1つ以上含む形態素ノードの数。
            ne_classes: dict
                固有名クラスをキー、そのクラスの地名語を1つ以上含む
                形態素ノードの数を値とする dict。

        Examples
        --------
        >>> import pygeonlp.api as api
        >>> api.init()
        >>> parser = api.parser.Parser()
        >>> api.filter.Filter.count_geowords(
        ...   parser.analyze_sentence('国会議事堂前まで歩きました。'))
        {'num_geowords': 1, 'num_addresses': 0,
            'ne_classes': {'鉄道施設/鉄道駅': 1, '鉄道施設': 1}}
        """
        num_geowords = 0
        num_addresses = 0
        ne_classes = {}
        for nodes in lattice:
            contains_geoword = False
            contains_address = False
            ne_list = []
            for node in nodes:
                if node.node_type == Node.NORMAL:
                    continue

                if node.node_type == Node.ADDRESS:
                    contains_address = True
                    continue

                if node.node_type == Node.GEOWORD:
                    contains_geoword = True
                    ne_class = node.prop.get('ne_class', None)
                    if ne_class:
                        ne_list.append(ne_class)

            if contains_geoword:
                num_geowords += 1

            if contains_address:
                num_addresses += 1

            for ne in set(ne_list):
                if ne in ne_classes:
                    ne_classes[ne] += 1
                else:
                    ne_classes[ne] = 1

                if '/' not in ne:
                    continue

                ne_basic = ne[0: ne.find('/')]
                if ne_basic in ne_classes:
                    ne_classes[ne_basic] += 1
                else:
                    ne_classes[ne_basic] = 1

        return {
            "num_geowords": num_geowords,
            "num_addresses": num_addresses,
            "ne_classes": ne_classes,
        }


class EntityClassFilter(Filter):
    """
    形態素ごとの候補ノード集合の中に、指定した固有名クラスと一致する地名語が
    含まれていれば、それ以外の候補は削除するフィルタ。
    テキストに含まれている地名語の固有名クラスが事前に分かっていて、
    そのクラスに一致する地名語だけを抽出したい場合などに利用します。

    候補ノード集合の中に一致する地名語が無い場合は候補リストをそのまま返します。
    """

    def __init__(self, ne_class, **kwargs):
        """
        Parameters
        ----------
        ne_class : str
            このフィルタで選択したい固有名クラスを表す正規表現文字列。
            例： r'国/?.*'
        """
        super().__init__(**kwargs)
        self.when_all_failed = 'convert_to_normal'
        self.re_ne_class = re.compile(ne_class)

    def filter_func(self, candidate, **kwargs):
        """
        Parameters
        ----------
        candidate : Node
            対象の形態素ノード

        Returns
        -------
        bool
            ne_class に一致する場合は True,
            一致しない場合は Falase
        """
        prop = candidate.prop
        if prop is None:
            # 非地名語ノードも削除する
            # 条件に該当する地名語ノードが無い場合は、
            # when_all_failed の指定により全てのノードが復元される
            return False

        ne_class = prop.get('ne_class', None)
        if ne_class is None:
            return False

        return self.re_ne_class.match(ne_class)


class GreedySearchFilter(Filter):
    """
    前後の地名語との関係を見て最もスコアが高くなる地名語を
    1 つだけ残し、それ以外は削除するフィルタ。

    ただし同点となる地名語が複数存在する場合は複数の候補を残します。
    """

    def __init__(self, scoring_class=None, scoring_options=None, **kwargs):
        """
        Parameters
        ----------
        scoring_class : class, optional
            パスのスコアとノード間のスコアを計算する関数を持つ
            スコアリングクラス。
            指定しない場合、``pygeonlp.api.scoring`` モジュール内の
            ``ScoringClass`` が利用されます。
        scoring_options : any, optional
            スコアリングクラスの初期化に渡すオプションパラメータ。
        """
        super().__init__(**kwargs)
        self.when_all_failed = 'return_all'

        if scoring_class is None:
            from .scoring import ScoringClass
            self.scorer = ScoringClass(scoring_options)
        else:
            self.scorer = scoring_class(scoring_options)

    def apply(self, input, **kwargs):
        """
        このフィルタは前後の形態素ノードを調べて地名語間の関係を計算するため、
        ``apply()`` をオーバーライドします。

        Parameters
        ----------
        input : list
            入力となるラティス表現。

        Returns
        -------
        list
            フィルタを適用したラティス表現。
        """
        output = [None] * len(input)

        # 地名語と住所を集めたリスト geowords_pile を作る
        # geowords_pile = [[pos, [geoword]]], pos はラティスでの位置
        geowords_pile = []
        n = 0
        while n < len(input):
            nodes = input[n]
            geowords = []
            for node in nodes:
                if node.node_type == Node.ADDRESS:
                    geowords.clear()
                    n += len(node.morphemes) - 1
                    break
                elif node.node_type == Node.GEOWORD:
                    geowords.append(node)

            if len(geowords) > 0:
                geowords_pile.append([n, geowords])

            n += 1

        for k, geowords in enumerate(geowords_pile):
            candidates = ",".join([x.simple() for x in geowords[1]])
            logger.debug("{}: {}".format(k, candidates))

        # ラティスの i 番目のノード集合を最適なものに絞り込む
        pile_index = 0
        for i in range(len(input)):
            if pile_index == len(geowords_pile) or \
                    i < geowords_pile[pile_index][0]:
                # 地名語・住所を含まない場合はそのまま
                output[i] = input[i]
                continue

            if len(input[i]) == 1:
                # 候補が一つしか無い場合もそのまま
                output[i] = input[i]
                pile_index += 1
                continue

            dt = 1  # チェックする前後の範囲
            best_nodes = input[i]
            while dt < len(geowords_pile):
                # pile_index - dt 番目の地名語集合を手がかりとする
                candidates = ",".join([x.simple() for x in input[i]])

                k = pile_index - dt
                if k >= 0:
                    hint_geowords = geowords_pile[k][1]
                    hints = ",".join([x.simple() for x in hint_geowords])
                    best_nodes = self._get_best(input[i], hint_geowords)
                    if len(best_nodes) == 1:
                        logger.debug("'{}' を選択（候補:{}, ヒント:{}）".format(
                            best_nodes[0].simple(), candidates, hints))
                        break

                k = pile_index + dt
                if k < len(geowords_pile):
                    hint_geowords = geowords_pile[k][1]
                    hints = ",".join([x.simple() for x in hint_geowords])
                    best_nodes = self._get_best(input[i], hint_geowords)
                    if len(best_nodes) == 1:
                        logger.debug("'{}' を選択（候補:{}, ヒント:{}）".format(
                            best_nodes[0].simple(), candidates, hints))
                        break

                dt += 1

            output[i] = best_nodes
            pile_index += 1

        return output

    def _get_best(self, nodes, hint_geowords):
        """
        形態素ノードリストの中から、前方・後方の地名語との関係が
        最も強いノードを返します。

        同じスコアになるノードが複数存在する場合は全て返します。

        Parameters
        ----------
        nodes : list
            形態素ノードのリスト
        hint_geowords : list
            地名語ノードまたは住所ノードを1つ以上含む、関係を調べる対象の
            ノードリスト

        Returns
        -------
        list
            関係が最も強い形態素ノード、または元のノードのリスト
        """
        scores = [[node, 0.0] for node in nodes]

        for i0 in range(len(nodes)):
            for i1 in range(len(hint_geowords)):
                scores[i0][1] += self.scorer.node_relation_score(
                    nodes[i0], hint_geowords[i1])

        results = []
        max_score = 0.0
        for score in scores:
            if score[1] > max_score:
                results = [score[0]]
                max_score = score[1]
            elif score[1] == max_score:
                results.append(score[0])

        return results


class InputBasedFilter(Filter):
    """
    入力によって対象とする地名語クラスやパス表現の選択ロジックを
    変更して、実時間で処理可能な候補に絞り込むフィルタ。
    """

    def __init__(self, **kwargs):
        """
        フィルタを初期化します。
        このクラスには指定するパラメータはありません。
        """
        super().__init__(**kwargs)

    def apply(self, lattice, **kwargs):
        """
        入力ベースフィルタを適用します。

        Parameters
        ----------
        lattice: list
            入力となるラティス表現。

        Returns
        -------
        list
            フィルタ適用後のラティス表現。

        Note
        ----
        時刻表や地震速報など、特定の地名語クラスを持つ地名が
        高い密度で出現する文字列を解析する場合、候補となるクラスを
        絞り込んで解析処理時間を短縮し、精度を向上させます。
        """
        filters = []

        search_plan = "exhaustive"
        gstat = Filter.count_geowords(lattice)
        ne_classes = gstat['ne_classes']
        num_geowords = gstat['num_geowords']

        if num_geowords >= 5 and \
           ne_classes.get('都道府県', 0) / num_geowords >= 0.75:
            # 都道府県リスト
            filters.append(EntityClassFilter(r'都道府県/?.*'))
            search_plan = "greedy"
        elif num_geowords >= 5 and \
                ne_classes.get('市区町村', 0) / num_geowords >= 0.75:
            # 市区町村リスト
            filters.append(EntityClassFilter(r'(都道府県|市区町村)/?.*'))
            search_plan = "greedy"
        elif num_geowords >= 5 and \
                ne_classes.get('鉄道施設', 0) / num_geowords >= 0.75:
            # 鉄道駅リスト（交通情報や時刻表など）
            filters.append(EntityClassFilter(r'(市区町村|鉄道施設)/?.*'))
            search_plan = "greedy"

        if search_plan == "greedy":
            # 地震速報や時刻表など同種の地名語が多数出現する場合
            filters.append(GreedySearchFilter())

        # 選択したフィルタを適用
        for f in filters:
            lattice = f(lattice)

        return lattice



import glob
import os

from setuptools import setup, Extension


# Target Files
BASE_DIR = os.path.dirname(__file__)
LIBGEONLP_DIR = os.path.join(BASE_DIR, 'libgeonlp')
LIBGEONLP_INCLUDE_DIR = os.path.join(LIBGEONLP_DIR, 'include')
LIBGEONLP_SOURCE_DIR = os.path.join(LIBGEONLP_DIR, 'lib')
LIBGEONLP_FILES = glob.glob(os.path.join(LIBGEONLP_SOURCE_DIR, '*.cpp'))
LIBGEONLP_HEADERS = glob.glob(os.path.join(LIBGEONLP_INCLUDE_DIR, '*.h'))
CPYGEONLP_FILES = glob.glob(os.path.join(LIBGEONLP_DIR, '*.cpp'))

# Other Variables
DESCRIPTION = __doc__
LONG_DESCRIPTION = open("README.md", "r", encoding="utf-8").read()
LICENSE = '2-Clause "Simplified" BSD License'
DATA_FILES = glob.glob(os.path.join(BASE_DIR, 'base_data', '*.*'))


def get_libgeonlp():
    """
    Returns the C++ extension `libgeonlp`.
    The logic for auto-detecting the libraries required for
    the extension is based on the following.
    https://stackoverflow.com/questions/28843765/setup-py-check-if-non-python-library-dependency-exists
    """
    from textwrap import dedent
    import tempfile
    import shutil
    import distutils.sysconfig
    import distutils.ccompiler
    from distutils.errors import LinkError

    boost_libs = ['boost_regex', 'boost_system', 'boost_filesystem']
    option_cflags = []

    # Search boost dirs
    boost_inc_dirs = None
    boost_lib_dirs = None
    extra_preargs = []
    for prefix in ('/usr', '/usr/local', '/opt', '/opt/homebrew'):
        if os.path.isdir(os.path.join(prefix, 'include/boost')):
            boost_inc_dirs = [os.path.join(prefix, 'include')]
            boost_lib_dirs = [os.path.join(prefix, 'lib')]
            if 'brew' in prefix:  # MacOSX + Homebrew
                option_cflags.append('-std=c++14')

            break

    # Write a temporary .c file to compile
    c_code = dedent("""
    #include <iostream>
    #include <string>
    #include <boost/regex.hpp>

    int main(int argc, char* argv[])
    {
      boost::regex r("air$");
      boost::smatch m;
      std::string pattern = "chair";

      if (boost::regex_search(pattern, m, r)) {
        std::cout << "match." << std::endl;
      } else {
        std::cout << "not match." << std::endl;
      }
      return 0;
    }
    """)

    tmp_dir = tempfile.mkdtemp(prefix='tmp_boost_')
    bin_file_name = os.path.join(tmp_dir, 'test_boost')
    file_name = bin_file_name + '.cpp'
    with open(file_name, 'w') as fp:
        fp.write(c_code)

    # Try compile and link it
    compiler = distutils.ccompiler.new_compiler()
    assert isinstance(compiler, distutils.ccompiler.CCompiler)
    distutils.sysconfig.customize_compiler(compiler)

    compiled = compiler.compile(
        sources=[file_name],
        include_dirs=boost_inc_dirs,
        output_dir='test_boost',
        extra_postargs=option_cflags,
    )

    libraries = None
    if not libraries:
        try:
            compiler.link_shared_lib(
                compiled,
                bin_file_name,
                library_dirs=boost_lib_dirs,
                libraries=boost_libs,
            )
        except LinkError:
            pass
        else:
            libraries = boost_libs

    if not libraries:
        boost_mt_libs = [lib + '-mt' for lib in boost_libs]
        try:
            compiler.link_shared_lib(
                compiled,
                bin_file_name,
                library_dirs=boost_lib_dirs,
                libraries=boost_mt_libs,
            )
        except LinkError:
            pass
        else:
            libraries = boost_mt_libs

    shutil.rmtree(tmp_dir)
    shutil.rmtree('test_boost')

    if not libraries:
        raise RuntimeError("Boost libraries is not installed.")

    libgeonlp = Extension(
        # see: https://setuptools.pypa.io/en/latest/deprecated/distutils/apiref.html#distutils.ccompiler.gen_preprocess_options  # noqa: E501
        'pygeonlp.capi',
        define_macros=[
            ('MAJOR_VERSION', '1'),
            ('MINOR_VERSION', '2'),
            ('REVISION', '4'),
        ],
        include_dirs=[LIBGEONLP_INCLUDE_DIR] + boost_inc_dirs,
        library_dirs=[LIBGEONLP_SOURCE_DIR] + boost_lib_dirs,
        sources=LIBGEONLP_FILES + CPYGEONLP_FILES,
        libraries=['sqlite3', 'mecab'] + libraries,
        extra_compile_args=option_cflags,
    )
    return libgeonlp


# Setup tools
setup(
    name='pygeonlp',
    version='1.2.4',
    description='A Python module for geotagging Japanese texts.',
    author='GeoNLP Project Team',
    author_email='geonlp@nii.ac.jp',
    url='https://github.com/geonlp-platform/pygeonlp',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    ext_modules=[get_libgeonlp()],
    packages=['pygeonlp.api', 'pygeonlp.webapi'],
    entry_points={
        'console_scripts': [
            'pygeonlp = pygeonlp.api.__main__:main'
        ]
    },
    python_requires='>=3.9',
    install_requires=[
        'requests',
        'chardet',
        'docopt',
        'lxml',
        'python-dateutil',
        'deprecated',
        'jageocoder',
        'geographiclib'
    ],
    data_files=[('pygeonlp_basedata', DATA_FILES)],
    license=LICENSE,
)



"""
このモジュールは、特定の時間範囲内に存在した地名語だけを抽出するといった
時間フィルタを実装しています。
"""

import datetime
from logging import getLogger
import re

from pygeonlp.api.filter import Filter, FilterError

logger = getLogger(__name__)


class TemporalFilter(Filter):
    """
    時間フィルタの基底クラス。

    時間フィルタは、形態素に対応する地名語ノード候補のうち、
    条件に一致するものを残し、一致しないものを削除します。
    全ての候補が条件に一致しない場合、品詞情報として固有名詞を持つ
    一般ノードを一つ作成して置き換えます。

    時間フィルタでは、地名語ノードの prop.valid_from および prop.valid_to を
    そのノードが存在する期間とします。

    - valid_from が設定されていない場合
        無限の過去から存在していたとみなします。
    - valid_to が設定されていない場合
        無限の未来まで存在しているとみなします。

    Attributes
    ----------
    duration : (datetime.date, datetime.date)
        期間の開始日と終了日のタプル。
    """

    def __init__(self, date_from, date_to=None, **kwargs):
        """
        Parameters
        ----------
        date_from : str, datetime.date, datetime.datetime
            期間の開始日時。
        date_to : str, datetime.date, datetime.datetime, optional
            期間の終了日時。

        Note
        ----
        date_to を省略した場合、 date_from が表す一時点が期間になります。
        たとえば TimeExistsFilter('2001-01-01') は、 2001年1月1日時点に
        存在していた地名語ノードを残すフィルタになります。
        """
        super().__init__()
        self.duration = self.__class__.get_duration_from_dates(
            date_from, date_to)
        self.when_all_failed = 'convert_to_normal'

    @classmethod
    def get_date_from_isostr(cls, datestr):
        """
        ISO フォーマット、または類似したフォーマットから date に相当する部分を取得します。

        Parameters
        ----------
        datestr : str
            日付または日時文字列。

        Returns
        -------
        datetime.date
            日付部分を表す date オブジェクト。

        """
        if datestr is None or datestr == '':
            return None

        re_date = r"(\d{4})[/\-\s](\d{1,2})[/\-\s](\d{1,2})"

        match = re.match(re_date, datestr)
        if not match:
            raise FilterError("'{}' は日付として解析できません".format(
                datestr))

        ymd = match.groups()
        date = datetime.date(int(ymd[0]), int(ymd[1]), int(ymd[2]))
        return date

    @classmethod
    def get_duration_from_dates(cls, datestr0, datestr1):
        """
        2つの日付文字列から期間を表す tuple を取得します。

        Parameters
        ----------
        datestr0 : str, datetime.date, datetime.datetime
            期間の開始日時を指定します。
        datestr1 : str, datetime.date, datetime.datetime
            期間の終了日時を指定します。

        Returns
        -------
        tuple
            期間の開始日時と終了日時を表す datetime.date オブジェクトの
            tuple を返します。
        """
        if isinstance(datestr0, (datetime.date, datetime.datetime)):
            date0 = datestr0
        elif isinstance(datestr0, str):
            date0 = cls.get_date_from_isostr(datestr0)
        else:
            raise TypeError(("1番目のパラメータの型は "
                             "str, datetime.date, datetime.datetime"
                             " のいずれかで指定してください。"))

        if not datestr1:
            date1 = date0
        else:
            if isinstance(datestr1, (datetime.date, datetime.datetime)):
                date1 = datestr1
            elif isinstance(datestr1, str):
                date1 = cls.get_date_from_isostr(datestr1)
            else:
                raise TypeError(("2番目のパラメータの型は "
                                 "str, datetime.date, datetime.datetime"
                                 " のいずれかで指定してください。"))

        if (date0 > date1):
            date0, date1 = date1, date0

        return (date0, date1)

    @classmethod
    def duration_from_candidate(cls, candidate):
        '''
        候補ノードの valid_from, valid_to から有効期間を表す
        tuple を取得する。
        valid_from, valid_to が含まれていない場合は None を返す。
        '''
        if candidate.prop is None:
            return None

        prop = candidate.prop
        date0 = cls.get_date_from_isostr(prop.get('valid_from', None))
        date1 = cls.get_date_from_isostr(prop.get('valid_to', None))

        return (date0, date1)


class TimeExistsFilter(TemporalFilter):
    """
    指定した時間範囲に存在していれば合格とするフィルタを作成します。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.temporal_filter import TimeExistsFilter
    >>> api.init()
    >>> tefilter = TimeExistsFilter('2000-01-01', '2001-01-01')
    >>> lattice = api.analyze('田無市と保谷市は2001年1月21日に合併して西東京市になりました。')
    >>> for nodes in lattice:
    ...     [x.simple() for x in nodes]
    ...
    ["田無市(GEOWORD:['東京都'])"]
    ['と(NORMAL)']
    ["保谷市(GEOWORD:['東京都'])"]
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ["西東京市(GEOWORD:['東京都'])"]
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']
    >>> lattice_filtered = tefilter(lattice)
    >>> # フィルタを適用すると「西東京市」が範囲外なので地名語候補から除外されます
    >>> for nodes in lattice_filtered:
    ...     [x.simple() for x in nodes]
    ...
    ["田無市(GEOWORD:['東京都'])"]
    ['と(NORMAL)']
    ["保谷市(GEOWORD:['東京都'])"]
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ['西東京市(NORMAL)']
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']

    Note
    ----
    このフィルタは TimeOverlapsFilter と同じです。
    """

    def filter_func(self, candidate):
        """
        時間範囲に存在していれば True を返します。

        Parameters
        ----------
        candidate : pygeonlp.api.node.Node
            検査対象となるノードオブジェクト。

        Returns
        -------
        bool
            フィルタの条件を満たせば True, 満たさなければ False を返します。
        """
        span = self.__class__.duration_from_candidate(candidate)
        if span is None:
            return True   # 期間を持たない候補は合格

        if span[0] and span[0] > self.duration[1]:
            return False

        if span[1] and span[1] < self.duration[0]:
            return False

        return True


class TimeBeforeFilter(TemporalFilter):
    """
    指定した時間範囲の開始時より前に存在していれば合格とするフィルタを返します。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.temporal_filter import TimeBeforeFilter
    >>> api.init()
    >>> tbfilter = TimeBeforeFilter('2000-01-01')
    >>> lattice = api.analyze('田無市と保谷市は2001年1月21日に合併して西東京市になりました。')
    >>> for nodes in lattice:
    ...     [x.simple() for x in nodes]
    ...
    ["田無市(GEOWORD:['東京都'])"]
    ['と(NORMAL)']
    ["保谷市(GEOWORD:['東京都'])"]
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ["西東京市(GEOWORD:['東京都'])"]
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']
    >>> lattice_filtered = tbfilter(lattice)
    >>> # フィルタを適用すると「西東京市」が範囲外なので地名語候補から除外されます
    >>> for nodes in lattice_filtered:
    ...     [x.simple() for x in nodes]
    ...
    ["田無市(GEOWORD:['東京都'])"]
    ['と(NORMAL)']
    ["保谷市(GEOWORD:['東京都'])"]
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ['西東京市(NORMAL)']
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']
    """

    def filter_func(self, candidate):
        """
        時間範囲の開始時より前に存在していれば True を返します。

        Parameters
        ----------
        candidate : pygeonlp.api.node.Node
            検査対象となるノードオブジェクト。

        Returns
        -------
        bool
            フィルタの条件を満たせば True, 満たさなければ False を返します。
        """
        span = self.__class__.duration_from_candidate(candidate)
        if span is None:
            return True   # 期間を持たない候補は合格

        if span[0] and span[0] > self.duration[0]:
            return False

        return True


class TimeAfterFilter(TemporalFilter):
    """
    指定した時間範囲の終了時より後に存在していれば合格とするフィルタを作成します。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.temporal_filter import TimeAfterFilter
    >>> api.init()
    >>> tafilter = TimeAfterFilter('2001-01-22')
    >>> lattice = api.analyze('田無市と保谷市は2001年1月21日に合併して西東京市になりました。')
    >>> for nodes in lattice:
    ...     [x.simple() for x in nodes]
    ...
    ["田無市(GEOWORD:['東京都'])"]
    ['と(NORMAL)']
    ["保谷市(GEOWORD:['東京都'])"]
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ["西東京市(GEOWORD:['東京都'])"]
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']
    >>> lattice_filtered = tafilter(lattice)
    >>> # フィルタを適用すると「田無市」「保谷市」が範囲外なので地名語候補から除外されます
    >>> for nodes in lattice_filtered:
    ...     [x.simple() for x in nodes]
    ...
    ['田無市(NORMAL)']
    ['と(NORMAL)']
    ['保谷市(NORMAL)']
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ["西東京市(GEOWORD:['東京都'])"]
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']
    """

    def filter_func(self, candidate):
        """
        時間範囲の開始時より前に存在していれば True を返します。

        Parameters
        ----------
        candidate : pygeonlp.api.node.Node
            検査対象となるノードオブジェクト。

        Returns
        -------
        bool
            フィルタの条件を満たせば True, 満たさなければ False を返します。
        """
        span = self.__class__.duration_from_candidate(candidate)
        if span is None:
            return True   # 期間を持たない候補は合格

        if span[1] and span[1] < self.duration[1]:
            return False

        return True


class TimeOverlapsFilter(TemporalFilter):
    """
    指定した時間範囲内に存在していれば合格とするフィルタを作成します。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.temporal_filter import TimeOverlapsFilter
    >>> api.init()
    >>> tofilter = TimeOverlapsFilter('2001-01-01', '2001-02-01')
    >>> lattice = api.analyze('田無市と保谷市は2001年1月21日に合併して西東京市になりました。')
    >>> for nodes in lattice:
    ...     [x.simple() for x in nodes]
    ...
    ["田無市(GEOWORD:['東京都'])"]
    ['と(NORMAL)']
    ["保谷市(GEOWORD:['東京都'])"]
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ["西東京市(GEOWORD:['東京都'])"]
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']
    >>> lattice_filtered = tofilter(lattice)
    >>> # フィルタを適用しても、「田無市」「保谷市」は2001年1月1日より後まで存在し、
    >>> # 「西東京市」は2001年2月1日より前から存在するので、いずれもそのまま残ります
    >>> for nodes in lattice_filtered:
    ...     [x.simple() for x in nodes]
    ...
    ["田無市(GEOWORD:['東京都'])"]
    ['と(NORMAL)']
    ["保谷市(GEOWORD:['東京都'])"]
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ["西東京市(GEOWORD:['東京都'])"]
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']

    Note
    ----
    このフィルタは TimeExistsFilter と同じです。
    """

    def filter_func(self, candidate):
        """
        時間範囲の開始時より後、または終了時より前に存在していれば True を返します。

        Parameters
        ----------
        candidate : pygeonlp.api.node.Node
            検査対象となるノードオブジェクト。

        Returns
        -------
        bool
            フィルタの条件を満たせば True, 満たさなければ False を返します。
        """
        span = self.__class__.duration_from_candidate(candidate)
        if span is None:
            return True   # 期間を持たない候補は合格

        if span[0] and span[0] > self.duration[1]:
            return False

        if span[1] and span[1] < self.duration[0]:
            return False

        return True


class TimeCoversFilter(TemporalFilter):
    """
    指定した時間範囲の開始時から終了時まで存在していれば合格とするフィルタを作成します。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.temporal_filter import TimeCoversFilter
    >>> api.init()
    >>> tcfilter = TimeCoversFilter('2001-01-01', '2001-02-01')
    >>> lattice = api.analyze('田無市と保谷市は2001年1月21日に合併して西東京市になりました。')
    >>> for nodes in lattice:
    ...     [x.simple() for x in nodes]
    ...
    ["田無市(GEOWORD:['東京都'])"]
    ['と(NORMAL)']
    ["保谷市(GEOWORD:['東京都'])"]
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ["西東京市(GEOWORD:['東京都'])"]
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']
    >>> lattice_filtered = tcfilter(lattice)
    >>> # フィルタを適用すると、「田無市」「保谷市」は2001年2月1日より前に消滅し、
    >>> # 「西東京市」は2001年1月1日より前から存在するので、地名語候補から除外されます。
    >>> for nodes in lattice_filtered:
    ...     [x.simple() for x in nodes]
    ...
    ['田無市(NORMAL)']
    ['と(NORMAL)']
    ['保谷市(NORMAL)']
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ['西東京市(NORMAL)']
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']
    """

    def filter_func(self, candidate):
        """
        時間範囲の開始時より前、かつ、終了時より後に存在していれば True を返します。

        Parameters
        ----------
        candidate : pygeonlp.api.node.Node
            検査対象となるノードオブジェクト。

        Returns
        -------
        bool
            フィルタの条件を満たせば True, 満たさなければ False を返します。
        """
        span = self.__class__.duration_from_candidate(candidate)
        if span is None:
            return True   # 期間を持たない候補は合格

        if span[0] and span[0] > self.duration[0]:
            return False

        if span[1] and span[1] < self.duration[1]:
            return False

        return True


class TimeContainsFilter(TemporalFilter):
    """
    指定した時間範囲の開始時と終了時の間だけ存在していれば合格とするフィルタを
    作成します。開始時より前に存在していたり、終了時より後に存在している場合は
    不合格になります。

    そのため、 valid_from, valid_to のいずれかが未指定の場合は不合格になります。

    Examples
    --------
    >>> import pygeonlp.api as api
    >>> from pygeonlp.api.temporal_filter import TimeContainsFilter
    >>> api.init()
    >>> tcfilter = TimeContainsFilter('2001-01-01', '2001-02-01')
    >>> lattice = api.analyze('田無市と保谷市は2001年1月21日に合併して西東京市になりました。')
    >>> for nodes in lattice:
    ...     [x.simple() for x in nodes]
    ...
    ["田無市(GEOWORD:['東京都'])"]
    ['と(NORMAL)']
    ["保谷市(GEOWORD:['東京都'])"]
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ["西東京市(GEOWORD:['東京都'])"]
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']
    >>> lattice_filtered = tcfilter(lattice)
    >>> # フィルタを適用すると、「田無市」「保谷市」は2001年1月1日より前にも存在し、
    >>> # 「西東京市」は2001年2月1日より後まで存在するので、地名語候補から除外されます。
    >>> for nodes in lattice_filtered:
    ...     [x.simple() for x in nodes]
    ...
    ['田無市(NORMAL)']
    ['と(NORMAL)']
    ['保谷市(NORMAL)']
    ['は(NORMAL)']
    ['2001(NORMAL)']
    ['年(NORMAL)']
    ['1(NORMAL)']
    ['月(NORMAL)']
    ['21(NORMAL)']
    ['日(NORMAL)']
    ['に(NORMAL)']
    ['合併(NORMAL)']
    ['し(NORMAL)']
    ['て(NORMAL)']
    ['西東京市(NORMAL)']
    ['に(NORMAL)']
    ['なり(NORMAL)']
    ['まし(NORMAL)']
    ['た(NORMAL)']
    ['。(NORMAL)']

    """

    def filter_func(self, candidate):
        """
        時間範囲の開始時と終了時の間に存在していれば True を返します。

        Parameters
        ----------
        candidate : pygeonlp.api.node.Node
            検査対象となるノードオブジェクト。

        Returns
        -------
        bool
            フィルタの条件を満たせば True, 満たさなければ False を返します。
        """
        span = self.__class__.duration_from_candidate(candidate)
        if span is None:
            return True   # 期間を持たない候補は合格

        if span[0] and span[0] >= self.duration[0] and \
           span[1] and span[1] <= self.duration[1]:
            return True

        return False
